// Scintilla source code edit control
// Messages.h - implementation of FOX->Scintilla messages
// The License.txt file describes the conditions under which this software may be distributed.
// SB: initial release; based on ScintillaWX by Robin Dunn.

#ifndef _MESSAGES_H
#define _MESSAGES_H

#define MAKELONG(Hi,Low) (((long)(unsigned)Hi << 16) | (long)(unsigned)Low)

//SB TODO: move the help funcs (those not using SendMsg()) to stc.cpp


//////////////////////////////////////////////////////////////////////////////
//// Text manipulation:

void AddText(const FXString& text) {
  SendMsg(SCI_ADDTEXT, text.length(), (long)text.text());
  SendMsg(SCI_SCROLLCARET, 0, 0);
}

// Add array of cells to document
void AddStyledText(const FXString& text) {
  SendMsg(2002, text.length(), (long)text.text());
}

// Insert string at a position
void InsertText(FXint pos, const FXString& text) {
  SendMsg(2003, pos, (long)text.text());
}

// Delete all text in the document
void ClearAll() {
  SendMsg(2004, 0, 0);
}

// Returns the character byte at the position
int GetCharAt(FXint pos) {
  return SendMsg(2007, pos, 0);
}

// Retrieve the text of the line containing the caret.
// Returns the index of the caret on the line.
FXString GetCurLine(FXint* linePos) {
	FXint len = LineLength(GetCaretLine());
	FXchar *buf; FXMALLOC(&buf, FXchar, len);
	FXint pos = SendMsg(2027, len, (long)buf);
	FXString text(buf);
	if (linePos)  *linePos = pos;
	return text;
}

// Retrieve the contents of a line.
FXString GetLine(FXint line) {
	FXint len = LineLength(line);
	FXchar *buf; FXMALLOC(&buf, FXchar, len);
  /* FXint pos = */ SendMsg(2153, line, (long)buf);
	FXString text(buf);
 return text;
}

// Retrieve the selected text.
FXString GetSelectedText() {
  FXint start, end;
  GetSelection(&start, &end);
  FXchar *buf; FXMALLOC(&buf, FXchar, end -start +1);
  SendMsg(2161, 0, (long)buf);
  FXString text(buf);
  return text;
}

// Retrieve a range of text.
FXString GetTextRange(int startPos, int endPos) {
  FXchar *buf; FXMALLOC(&buf, FXchar, endPos -startPos +1);
  TextRange tr;
	tr.lpstrText = buf;
	tr.chrg.cpMin = startPos;
	tr.chrg.cpMax = endPos;
	SendMsg(2162, 0, (long)&tr);
	FXString text(buf);
	return text;
}

// Retrieve a buffer of cells.
FXString GetStyledText(int startPos, int endPos) {
  FXchar *buf; FXMALLOC(&buf, FXchar, endPos -startPos +1);
  TextRange tr;
	tr.lpstrText = buf;
	tr.chrg.cpMin = startPos;
	tr.chrg.cpMax = endPos;
	SendMsg(2015, 0, (long)&tr);
	FXString text(buf);
	return text;

}

// Replace the contents of the document with the argument text.
void SetText(const FXString& text) {
  SendMsg(2181, 0, (long)text.text());
}

// Retrieve all the text in the document.
FXString GetText() {
	FXint len = GetTextLength();
	FXchar *buf; FXMALLOC(&buf, FXchar, len);
  SendMsg(2182, len, (long)buf);
	FXString text(buf);
 return text;
}



//////////////////////////////////////////////////////////////////////////////
//// Selection

// Select all the text in the document.
void SelectAll() {
  SendMsg(2013, 0, 0);
}

// Returns the position of the opposite end of the selection to the caret
int GetAnchor() {
  return SendMsg(2009, 0, 0);
}

// Set the selection anchor to a position. The anchor is the opposite
// end of the selection from the caret.
void SetAnchor(FXint posAnchor) {
  SendMsg(2026, posAnchor, 0);
}

// Set the foreground colour of the selection and whether to use this setting.
void SetSelForeground(FXbool useSetting, const FXColor& fore) {
  SendMsg(2067, useSetting, (long)fore);
}

// Set the background colour of the selection and whether to use this setting.
void SetSelBackground(FXbool useSetting, const FXColor& back) {
  SendMsg(2068, useSetting, (long)back);
}

// Sets the position that starts the selection - this becomes the anchor.
void SetSelectionStart(FXint pos) {
  SendMsg(2142, pos, 0);
}

// Returns the position at the start of the selection.
FXint GetSelectionStart() {
  return SendMsg(2143, 0, 0);
}

// Sets the position that ends the selection - this becomes the currentPosition.
void SetSelectionEnd(int pos) {
  SendMsg(2144, pos, 0);
}

// Returns the position at the end of the selection.
FXint GetSelectionEnd() {
  return SendMsg(2145, 0, 0);
}

// Select a range of text.
void SetSelection(int start, int end) {
  SendMsg(2160, start, end);
}

// Draw the selection in normal style or with selection highlighted.
void HideSelection(FXbool normal) {
  SendMsg(2163, normal, 0);
}

// Replace the selected text with the argument text.
void ReplaceSelection(const FXString& text) {
  SendMsg(2170, 0, (long)text.text());
}

// Delete (Clear) the selection.
void ClearSelection() {
  SendMsg(2180, 0, 0);
}

// Is the selection a rectangular. The alternative is the more common stream selection.
FXbool SelectionIsRectangle() {
  return SendMsg(2372, 0, 0) != 0;
}

// wx:
// Retrieve the start and end positions of the current selection.
void GetSelection(FXint* startPos, FXint* endPos) {
  if (startPos != NULL) *startPos = SendMsg(SCI_GETSELECTIONSTART);
  if (endPos != NULL) *endPos = SendMsg(SCI_GETSELECTIONEND);
}



//////////////////////////////////////////////////////////////////////////////
//// Margins {1, 2, 3}

// Set a margin to be either numeric or symbolic.
void SetMarginType(FXint margin, FXint marginType) {
  SendMsg(SCI_SETMARGINTYPEN, margin, marginType);
  //update();
}

// Set a mask that determines which markers are displayed in a margin.
void SetMarginMask(FXint margin, FXint marginMask) {
  SendMsg(SCI_SETMARGINMASKN, margin, marginMask);
  //update();
}


// Retrieve the type of a margin.
FXint GetMarginType(FXint margin) {
  return SendMsg(2241, margin, 0);
}

// Set the width of a margin to a width expressed in pixels.
void SetMarginWidth(FXint margin, FXint pixelWidth) {
  SendMsg(2242, margin, pixelWidth);
}

// Retrieve the width of a margin in pixels.
FXint GetMarginWidth(FXint margin) {
  return SendMsg(2243, margin, 0);
}


// Retrieve the marker mask of a margin.
FXint GetMarginMask(FXint margin) {
  return SendMsg(2245, margin, 0);
}

// Make a margin sensitive or insensitive to mouse clicks.
void SetMarginSensitive(FXint margin, FXbool sensitive) {
  SendMsg(2246, margin, sensitive);
}

// Retrieve the mouse click sensitivity of a margin.
FXbool GetMarginSensitive(FXint margin) {
  return SendMsg(2247, margin, 0) != 0;
}

// Sets the size in pixels of the left margin.
void SetMarginLeft(FXint width) {
  SendMsg(2155, 0, width);
}

// Returns the size in pixels of the left margin.
FXint GetMarginLeft() {
  return SendMsg(2156, 0, 0);
}

// Sets the size in pixels of the right margin.
void SetMarginRight(FXint width) {
  SendMsg(2157, 0, width);
}

// Returns the size in pixels of the right margin.
int GetMarginRight() {
  return SendMsg(2158, 0, 0);
}



//////////////////////////////////////////////////////////////////////////////
//// Styling

// Set the lexing language of the document.
void SetLexer(FXint lexer) {
  SendMsg(4001, lexer, 0);
}

// Retrieve the lexing language of the document.
FXint GetLexer() {
  return SendMsg(4002, 0, 0);
}

// Colourise a segment of the document using the current lexing language.
void Colourise(FXint start, FXint end) {
  SendMsg(4003, start, end);
}

// Set the size of characters of a style.
void StyleSetSize(FXint style, FXint sizePoints) {
  SendMsg(SCI_STYLESETSIZE, style, sizePoints);
}

// Set the font of a style.
void StyleSetFace(FXint style, const FXString& fontName) {
  SendMsg(SCI_STYLESETFONT, style, (long)fontName.text());
}

// Set all style bytes to 0, remove all folding information
void ClearDocumentStyle() {
  SendMsg(2005, 0, 0);
}

// Returns the style byte at the position
int GetStyleAt(int pos) {
  return SendMsg(2010, pos, 0);
}

// Retrieve the position of the last correctly styled character.
FXint GetEndStyled() {
  return SendMsg(2028, 0, 0);
}

// Set the current styling position to pos and the styling mask to mask.
// The styling mask can be used to protect some bits in each styling byte from
// modification.
void StartStyling(FXint pos, FXint mask) {
  SendMsg(2032, pos, mask);
}

// Change style from current styling position for length characters to a style
// and move the current styling position to after this newly styled segment.
void SetStyling(FXint length, FXint style) {
  SendMsg(2033, length, style);
}

// Clear all the styles and make equivalent to the global default style.
void StyleClearAll() {
  SendMsg(2050, 0, 0);
}

// Set the foreground colour of a style.
void StyleSetForeground(FXint style, const FXColor& fore) {
  SendMsg(2051, style, (long)fore);
}

// Set the background colour of a style.
void StyleSetBackground(FXint style, const FXColor& back) {
  SendMsg(2052, style, (long)back);
}

// Set a style to be bold or not.
void StyleSetBold(FXint style, FXbool bold) {
  SendMsg(2053, style, bold);
}

// Set a style to be italic or not.
void StyleSetItalic(FXint style, FXbool italic) {
  SendMsg(2054, style, italic);
}

// Set the font of a style.
void StyleSetFaceName(FXint style, const FXString& fontName) {
  SendMsg(2056, style, (long)fontName.text());
}

// Set a style to have its end of line filled or not.
void StyleSetEOLFilled(FXint style, FXbool filled) {
  SendMsg(2057, style, filled);
}

// Reset the default style to its state at startup
void StyleResetDefault() {
  SendMsg(2058, 0, 0);
}

// Set a style to be underlined or not.
void StyleSetUnderline(FXint style, FXbool underline) {
  SendMsg(2059, style, underline);
}

// Set the styles for a segment of the document.
void SetStyleBytes(FXint length, char* styleBytes) {
  SendMsg(2073, length, (long)styleBytes);
}

// Set a style to be visible or not.
void StyleSetVisible(FXint style, FXbool visible) {
  SendMsg(2074, style, visible);
}

// Divide each styling byte into lexical class bits (default:5) and indicator
// bits (default:3). If a lexer requires more than 32 lexical states, then this
// is used to expand the possible states.
void SetStyleBits(FXint bits) {
  SendMsg(2090, bits, 0);
}

// Retrieve number of bits in style bytes used to hold the lexical state.
FXint GetStyleBits() {
  return SendMsg(2091, 0, 0);
}

// Used to hold extra styling information for each line.
void SetLineState(FXint line, FXint state) {
  SendMsg(2092, line, state);
}

// Retrieve the extra styling information for a line.
FXint GetLineState(FXint line) {
  return SendMsg(2093, line, 0);
}

// Retrieve the last line number that has line state.
FXint GetMaxLineState() {
  return SendMsg(2094, 0, 0);
}

// Set up a value that may be used by a lexer for some optional feature.
void SetProperty(const FXString& key, const FXString& value) {
  SendMsg(4004, (long)key.text(), (long)value.text());
}

// Set up the key words used by the lexer.
void SetKeyWords(FXint keywordSet, const FXString& keyWords) {
  SendMsg(4005, keywordSet, (long)keyWords.text());
}

//wx:
// more convenient funcs in wx:stc.cpp



//////////////////////////////////////////////////////////////////////////////
//// Information

// Returns the number of lines in the document. There is always at least one.
FXint GetLineCount() {
  return SendMsg(2154, 0, 0);
}

// How many characters are on a line, not including end of line characters.
FXint LineLength(FXint line) {
  return SendMsg(2350, line, 0);
}

// Retrieve the line containing a position.
FXint LineFromPosition(FXint pos) {
  return SendMsg(2166, pos, 0);
}

// Find the position from a point within the window.
FXint PositionFromPoint(FXPoint pt) {
  return SendMsg(2022, pt.x, pt.y);
}

// The number of characters in the document
FXint GetLength() {
  return SendMsg(2006, 0, 0);
}

// Returns the position of the caret
FXint GetCurrentPos() {
  return SendMsg(2008, 0, 0);
}

// Retrieve the column number of a position, taking tab width into account.
FXint GetColumn(FXint pos) {
  return SendMsg(2129, pos, 0);
}

// Get the position after the last visible characters on a line.
FXint GetLineEndPosition(FXint line) {
  return SendMsg(2136, line, 0);
}

// Get the code page used to interpret the bytes of the document as characters.
FXint GetCodePage() {
  return SendMsg(2137, 0, 0);
}

// In read-only mode?
FXbool GetReadOnly() {
  return SendMsg(2140, 0, 0) != 0;
}

// Retrieve the line at the top of the display.
FXint GetFirstVisibleLine() {
  return SendMsg(2152, 0, 0);
}

// Is the document different from when it was last saved?
FXbool GetModify() {
  return SendMsg(2159, 0, 0) != 0;
}

// Retrieve the position at the start of a line.
FXint PositionFromLine(FXint line) {
  return SendMsg(2167, line, 0);
}

// Retrieve the number of characters in the document.
FXint GetTextLength() {
  return SendMsg(2183, 0, 0);
}

// Find the display line of a document line taking hidden lines into account.
FXint VisibleFromDocLine(FXint line) {
  return SendMsg(2220, line, 0);
}

// Find the document line of a display line taking hidden lines into account.
FXint DocLineFromVisible(FXint lineDisplay) {
  return SendMsg(2221, lineDisplay, 0);
}

// Retrieves the number of lines completely visible.
FXint LinesOnScreen() {
  return SendMsg(2370, 0, 0);
}

// Is a line visible?
FXbool GetLineVisible(FXint line) {
  return SendMsg(2228, line, 0) != 0;
}

// SB:
FXint GetCaretLine() {
  return LineFromPosition(GetCurrentPos());
}

//wx:
// Retrieve the point in the window where a position is displayed.
FXPoint PointFromPosition(FXint pos) {
  FXPoint pt;
  pt.x = SendMsg(SCI_POINTXFROMPOSITION, 0, pos);
  pt.y = SendMsg(SCI_POINTYFROMPOSITION, 0, pos);
  return pt;
}

//SB
FXbool IsPointInSelection(FXPoint fxpt) {
  return m_sfx->IsPointInSelection(fxpt);
}

FXbool IsPointInSelMargin(FXPoint fxpt) {
  return m_sfx->IsPointInSelMargin(fxpt);
}



//////////////////////////////////////////////////////////////////////////////
//// Movement

// Scroll horizontally and vertically.
void LineScroll(FXint columns, FXint lines) {
  SendMsg(2168, columns, lines);
}

// Set caret to a position and ensure it is visible.
void GotoPos(FXint pos) {
  SendMsg(2025, pos, 0);
}

// Set caret to start of a line and ensure it is visible.
void GotoLine(FXint line) {
  SendMsg(2024, line, 0);
}

// Ensure a particular line is visible by expanding any header line hiding it.
void EnsureVisible(FXint line) {
  SendMsg(2232, line, 0);
}

// Make a range of lines visible.
void ShowLines(FXint lineStart, FXint lineEnd) {
  SendMsg(2226, lineStart, lineEnd);
}

// Make a range of lines invisible.
void HideLines(FXint lineStart, FXint lineEnd) {
  SendMsg(2227, lineStart, lineEnd);
}

// wx: [TODO review]
// Scroll enough to make the given line visible
void ScrollToLine(FXint line) {
  m_sfx->DoScrollToLine(line);
}

// wx: [TODO review]
// Scroll enough to make the given column visible
void ScrollToColumn(FXint column) {
  m_sfx->DoScrollToColumn(column);
}



//////////////////////////////////////////////////////////////////////////////
//// Undo/Redo

// Redoes the next action on the undo history
void Redo() {
  SendMsg(2011, 0, 0);
}

// Choose between collecting actions into the undo
// history and discarding them.
void SetUndoCollection(FXbool collectUndo) {
  SendMsg(2012, collectUndo, 0);
}

// Remember the current position in the undo history as the position
// at which the document was saved.
void SetSavePoint() {
  SendMsg(2014, 0, 0);
}

// Are there any redoable actions in the undo history.
FXbool CanRedo() {
  return SendMsg(2016, 0, 0) != 0;
}

// Is undo history being collected?
FXbool GetUndoCollection() {
  return SendMsg(2019, 0, 0) != 0;
}

// Start a sequence of actions that is undone and redone as a unit.
// May be nested.
void BeginUndoAction() {
  SendMsg(2078, 0, 0);
}

// End a sequence of actions that is undone and redone as a unit.
void EndUndoAction() {
  SendMsg(2079, 0, 0);
}

// Are there any undoable actions in the undo history.
FXbool CanUndo() {
  return SendMsg(2174, 0, 0) != 0;
}

// Delete the undo history.
void EmptyUndoBuffer() {
  SendMsg(2175, 0, 0);
}

// Undo one action in the undo history.
void Undo() {
  SendMsg(2176, 0, 0);
}



//////////////////////////////////////////////////////////////////////////////
//// Markers

// Retrieve the line number at which a particular marker is located
int MarkerLineFromHandle(int handle) {
    return SendMsg(2017, handle, 0);
}

// Delete a marker.
void MarkerDeleteHandle(int handle) {
    SendMsg(2018, handle, 0);
}

// Set the symbol used for a particular marker number.
void MarkerDefine(int markerNumber, int markerSymbol) {
  SendMsg(2040, markerNumber, markerSymbol);
}

// Set the foreground colour used for a particular marker number.
void MarkerSetForeground(int markerNumber, const FXColor& fore) {
  //SendMsg(2041, markerNumber, wxColourAsLong(fore));
  SendMsg(2041, markerNumber, (long)fore);
}

// Set the background colour used for a particular marker number.
void MarkerSetBackground(int markerNumber, const FXColor& back) {
  //SendMsg(2042, markerNumber, wxColourAsLong(back));
  SendMsg(2042, markerNumber, (long)back);
}

// Add a marker to a line.
void MarkerAdd(int line, int markerNumber) {
  SendMsg(2043, line, markerNumber);
}

// Delete a marker from a line
void MarkerDelete(int line, int markerNumber) {
  SendMsg(2044, line, markerNumber);
}

// Delete all markers with a particular number from all lines
void MarkerDeleteAll(int markerNumber) {
  SendMsg(2045, markerNumber, 0);
}

// Get a bit mask of all the markers set on a line.
int MarkerGet(int line) {
  return SendMsg(2046, line, 0);
}

// Find the next line after lineStart that includes a marker in mask.
int MarkerNext(int lineStart, int markerMask) {
  return SendMsg(2047, lineStart, markerMask);
}

// Find the previous line before lineStart that includes a marker in mask.
int MarkerPrevious(int lineStart, int markerMask) {
  return SendMsg(2048, lineStart, markerMask);
}


//// Caret

// Set the foreground colour of the caret.
void SetCaretForeground(const FXColor& fore) {
  SendMsg(2069, (long)fore, 0);
}

// Get the time in milliseconds that the caret is on and off.
int GetCaretPeriod() {
  return SendMsg(2075, 0, 0);
}

// Get the time in milliseconds that the caret is on and off. 0 = steady on.
void SetCaretPeriod(int periodMilliseconds) {
  SendMsg(2076, periodMilliseconds, 0);
}

// Get the foreground colour of the caret.
FXColor GetCaretForeground() {
  FXlong lcol = SendMsg(2138, 0, 0);
  FXColor fxcol(lcol);  //SB ?
  //fxcol = FXRGB(lcol & 0xff, (lcol >> 8) & 0xff, (lcol >> 16) & 0xff);    //SB ?
  return fxcol;
}

// Sets the position of the caret.
void SetCurrentPos(int pos) {
  SendMsg(2141, pos, 0);
}

// Ensure the caret is visible.
void EnsureCaretVisible() {
  SendMsg(2169, 0, 0);
}

// Set the way the line the caret is on is kept visible.
void SetCaretPolicy(int caretPolicy, int caretSlop) {
  SendMsg(2369, caretPolicy, caretSlop);
}


//// Indicators

// Set an indicator to plain, squiggle or TT.
void IndicatorSetStyle(FXint indic, FXint style) {
  SendMsg(2080, indic, style);
}

// Retrieve the style of an indicator.
FXint IndicatorGetStyle(FXint indic) {
  return SendMsg(2081, indic, 0);
}

// Set the foreground colour of an indicator.
void IndicatorSetForeground(FXint indic, const FXColor& fore) {
  SendMsg(2082, indic, (long)fore);
}

// Retrieve the foreground colour of an indicator.
FXColor IndicatorGetForeground(FXint indic) {
  FXlong lcol = SendMsg(2083, indic, 0);
  FXColor fxcol(lcol);
  return fxcol;  //SB ?
}


//// Auto completion

// Display a auto-completion list.
// The lenEntered parameter indicates how many characters before
// the caret should be used to provide context.
void AutoCompShow(FXint lenEntered, const FXString& itemList) {
  SendMsg(2100, lenEntered, (long)itemList.text());
}

// Remove the auto-completion list from the screen.
void AutoCompCancel() {
  SendMsg(2101, 0, 0);
}

// Is there an auto-completion list visible?
bool AutoCompActive() {
  return SendMsg(2102, 0, 0) != 0;
}

// Retrieve the position of the caret when the auto-completion list was
// displayed.
FXint AutoCompPosStart() {
  return SendMsg(2103, 0, 0);
}

// User has selected an item so remove the list and insert the selection.
void AutoCompComplete() {
  SendMsg(2104, 0, 0);
}

// Define a set of character that when typed cancel the auto-completion list.
void AutoCompStops(const FXString& characterSet) {
  SendMsg(2105, 0, (long)characterSet.text());
}

// Change the separator character in the string setting up an auto-completion
// list. Default is space but can be changed if items contain space.
void AutoCompSetSeparator(int separatorCharacter) {
  SendMsg(2106, separatorCharacter, 0);
}

// Retrieve the auto-completion list separator character.
FXint AutoCompGetSeparator() {
  return SendMsg(2107, 0, 0);
}

// Select the item in the auto-completion list that starts with a string.
void AutoCompSelect(const FXString& text) {
  SendMsg(2108, 0, (long)text.text());
}

// Should the auto-completion list be cancelled if the user backspaces to a
// position before where the box was created.
void AutoCompSetCancelAtStart(bool cancel) {
  SendMsg(2110, cancel, 0);
}

// Retrieve whether auto-completion cancelled by backspacing before start.
bool AutoCompGetCancelAtStart() {
  return SendMsg(2111, 0, 0) != 0;
}

// Define a set of character that when typed fills up the selected word.
void AutoCompSetFillUps(const FXString& characterSet) {
  SendMsg(2112, 0, (long)characterSet.text());
}

// Should a single item auto-completion list automatically choose the item.
void AutoCompSetChooseSingle(bool chooseSingle) {
  SendMsg(2113, chooseSingle, 0);
}

// Retrieve whether a single item auto-completion list automatically choose the item.
bool AutoCompGetChooseSingle() {
  return SendMsg(2114, 0, 0) != 0;
}

// Set whether case is significant when performing auto-completion searches.
void AutoCompSetIgnoreCase(bool ignoreCase) {
  SendMsg(2115, ignoreCase, 0);
}

// Retrieve state of ignore case flag.
bool AutoCompGetIgnoreCase() {
  return SendMsg(2116, 0, 0) != 0;
}


//// Intendation

// Set the number of spaces used for one level of indentation.
void SetIndent(FXint indentSize) {
  SendMsg(2122, indentSize, 0);
}

// Retrieve indentation size.
int GetIndent() {
  return SendMsg(2123, 0, 0);
}

// Indentation will only use space characters if useTabs is false, otherwise
// it will use a combination of tabs and spaces.
void SetUseTabs(bool useTabs) {
  SendMsg(2124, useTabs, 0);
}

// Retrieve whether tabs will be used in indentation.
bool GetUseTabs() {
  return SendMsg(2125, 0, 0) != 0;
}

// Change the indentation of a line to a number of columns.
void SetLineIndentation(FXint line, FXint indentSize) {
  SendMsg(2126, line, indentSize);
}

// Retrieve the number of columns that a line is indented.
int GetLineIndentation(FXint line) {
  return SendMsg(2127, line, 0);
}

// Retrieve the position before the first non indentation character on a line.
int GetLineIndentPosition(FXint line) {
  return SendMsg(2128, line, 0);
}

// Show or hide indentation guides.
void SetIndentationGuides(bool show) {
  SendMsg(2132, show, 0);
}

// Are the indentation guides visible?
bool GetIndentationGuides() {
  return SendMsg(2133, 0, 0) != 0;
}

// Set the highlighted indentation guide column.
// 0 = no highlighted guide.
void SetHighlightGuide(FXint column) {
  SendMsg(2134, column, 0);
}

// Get the highlighted indentation guide column.
int GetHighlightGuide() {
  return SendMsg(2135, 0, 0);
}



//// Printing

// Sets the print magnification added to the point size of each style for printing.
void SetPrintMagnification(FXint magnification) {
  SendMsg(2146, magnification, 0);
}

// Returns the print magnification.
int GetPrintMagnification() {
  return SendMsg(2147, 0, 0);
}

// Modify colours when printing for clearer printed text.
void SetPrintColourMode(FXint mode) {
  SendMsg(2148, mode, 0);
}

// Returns the print colour mode.
int GetPrintColourMode() {
  return SendMsg(2149, 0, 0);
}


//// Search/Replace

// Find some text in the document.
FXint FindText(FXint minPos, FXint maxPos,
               const FXString& text,
               bool caseSensitive, bool wholeWord) {
  TextToFind  ft;
  int flags = 0;
  flags |= caseSensitive ? SCFIND_MATCHCASE : 0;
  flags |= wholeWord     ? SCFIND_WHOLEWORD : 0;
  ft.chrg.cpMin = minPos;
  ft.chrg.cpMax = maxPos;
  ft.lpstrText = (char*)text.text();
  return SendMsg(2150, flags, (long)&ft);
}

// Sets the current caret position to be the search anchor.
void SearchAnchor() {
  SendMsg(2366, 0, 0);
}

// Find some text starting at the search anchor.
int SearchNext(FXint flags, const FXString& text) {
  return SendMsg(2367, flags, (long)text.text());
}

// Find some text starting at the search anchor and moving backwards.
int SearchPrev(FXint flags, const FXString& text) {
  return SendMsg(2368, flags, (long)text.text());
}


//// Cut/Copy/Paste

// Will a paste succeed?
bool CanPasteSel() {
  return SendMsg(2173, 0, 0) != 0;
}

// Cut the selection to the clipboard.
void CutSel() {
  SendMsg(2177, 0, 0);
}

// Copy the selection to the clipboard.
void CopySel() {
  SendMsg(2178, 0, 0);
}

// Paste the contents of the clipboard into the document replacing the selection.
void PasteSel() {
  SendMsg(2179, 0, 0);
}


//// Call Tips

// Show a call tip containing a definition near position pos.
void CallTipShow(FXint pos, const FXString& definition) {
  SendMsg(2200, pos, (long)definition.text());
}

// Remove the call tip from the screen.
void CallTipCancel() {
  SendMsg(2201, 0, 0);
}

// Is there an active call tip?
bool CallTipActive() {
  return SendMsg(2202, 0, 0) != 0;
}

// Retrieve the position where the caret was before displaying the call tip.
int CallTipPosAtStart() {
  return SendMsg(2203, 0, 0);
}

// Highlight a segment of the definition.
void CallTipSetHighlight(FXint start, FXint end) {
  SendMsg(2204, start, end);
}

// Set the background colour for the call tip.
void CallTipSetBackground(const FXColor& back) {
  SendMsg(2205, (long)back, 0);
}


//// Folding

// Set the fold level of a line.
// This encodes an integer level along with flags indicating whether the
// line is a header and whether it is effectively white space.
void SetFoldLevel(FXint line, FXint level) {
  SendMsg(2222, line, level);
}

// Retrieve the fold level of a line.
int GetFoldLevel(FXint line) {
  return SendMsg(2223, line, 0);
}

// Find the last child line of a header line.
int GetLastChild(FXint line, FXint level) {
  return SendMsg(2224, line, level);
}

// Find the parent line of a child line.
int GetFoldParent(FXint line) {
  return SendMsg(2225, line, 0);
}

// Show the children of a header line.
void SetFoldExpanded(FXint line, bool expanded) {
  SendMsg(2229, line, expanded);
}

// Is a header line expanded?
bool GetFoldExpanded(FXint line) {
  return SendMsg(2230, line, 0) != 0;
}

// Switch a header line between expanded and contracted.
void ToggleFold(FXint line) {
  SendMsg(2231, line, 0);
}

// Set some debugging options for folding
void SetFoldFlags(FXint flags) {
  SendMsg(2233, flags, 0);
}


//// EOL/Edge

// Convert all line endings in the document to use the current mode.
void ConvertEOLs() {
  SendMsg(2029, 0, 0);
}

// Retrieve the current end of line mode - one of CRLF, CR, or LF.
int GetEOLMode() {
  return SendMsg(2030, 0, 0);
}

// Set the current end of line mode.
void SetEOLMode(FXint eolMode) {
  SendMsg(2031, eolMode, 0);
}

// Are the end of line characters visible.
bool GetViewEOL() {
  return SendMsg(2355, 0, 0) != 0;
}

// Make the end of line characters visible or invisible
void SetViewEOL(bool visible) {
  SendMsg(2356, visible, 0);
}

// Retrieve the column number which text should be kept within.
int GetEdgeColumn() {
  return SendMsg(2360, 0, 0);
}

// Set the column number of the edge.
// If text goes past the edge then it is highlighted.
void SetEdgeColumn(FXint column) {
  SendMsg(2361, column, 0);
}

// Retrieve the edge highlight mode.
int GetEdgeMode() {
  return SendMsg(2362, 0, 0);
}

// The edge may be displayed by a line (EDGE_LINE) or by highlighting text that
// goes beyond it (EDGE_BACKGROUND) or not displayed at all (EDGE_NONE).
void SetEdgeMode(FXint mode) {
  SendMsg(2363, mode, 0);
}

// Retrieve the colour used in edge indication.
FXColor GetEdgeColour() {
  FXlong lcol = SendMsg(2364, 0, 0);
  FXColor fxcol(lcol);
  return fxcol;
}

// Change the colour used in edge indication.
void SetEdgeColour(const FXColor& edgeColour) {
    SendMsg(2365, (long)edgeColour, 0);
}


//// Document

// Retrieve a pointer to the document object.
void* GetDocPointer() {
  return (void*)SendMsg(2357);
}

// Change the document object used.
void SetDocPointer(void* docPointer) {
  SendMsg(2358, (long)docPointer);
}

// Set which document modification events are sent to the container.
void SetModEventMask(int mask) {
    SendMsg(2359, mask, 0);
}

// Create a new document object.
// Starts with reference count of 1 and not selected into editor.
void* CreateDocument() {
  return (void*)SendMsg(2375);
}

// Extend life of document.
void AddRefDocument(void* docPointer) {
  SendMsg(2376, (long)docPointer);
}

// Release a reference to the document, deleting document if it fades to black.
void ReleaseDocument(void* docPointer) {
  SendMsg(2377, (long)docPointer);
}

// Get which document modification events are sent to the container.
int GetModEventMask() {
  return SendMsg(2378, 0, 0);
}


//// Macros

// Start notifying the container of all key presses and commands.
void StartRecord() {
  SendMsg(3001, 0, 0);
}

// Stop notifying the container of all key presses and commands.
void StopRecord() {
  SendMsg(3002, 0, 0);
}


//// Brace Highlight

// Highlight the characters at two positions.
void BraceHighlight(FXint pos1, FXint pos2) {
  SendMsg(2351, pos1, pos2);
}

// Highlight the character at a position indicating there is no matching brace.
void BraceBadLight(FXint pos) {
  SendMsg(2352, pos, 0);
}

// Find the position of a matching brace or INVALID_POSITION if no match.
int BraceMatch(FXint pos) {
  return SendMsg(2353, pos, 0);
}


//// Spaces/Tabs

// Are white space characters currently visible?
// Returns one of SCWS_* constants.
int GetViewWhiteSpace() {
  return SendMsg(2020, 0, 0);
}

// Make white space characters invisible, always visible or visible outside indentation.
void SetViewWhiteSpace(FXint viewWS) {
  SendMsg(2021, viewWS, 0);
}

// Change the visible size of a tab to be a multiple of the width of a space
// character.
void SetTabWidth(FXint tabWidth) {
  SendMsg(2036, tabWidth, 0);
}

// Retrieve the visible size of a tab.
int GetTabWidth() {
  return SendMsg(2121, 0, 0);
}


//// Key Binding

// When key+modifier combination km is pressed perform msg.
void CmdKeyAssign(int key, int modifiers, int cmd) {
  SendMsg(2070, MAKELONG(key, modifiers), cmd);
}

// When key+modifier combination km do nothing.
void CmdKeyClear(int key, int modifiers) {
  SendMsg(2071, MAKELONG(key, modifiers));
}

// Drop all key mappings.
void CmdKeyClearAll() {
  SendMsg(2072, 0, 0);
}

//Todo...
// Perform one of the operations defined by the wxSTC_CMD_* constants.
void CmdKeyExecute(int cmd) {
  SendMsg(cmd);
}


//// Misc

// Is drawing done first into a buffer or direct to the screen.
bool GetBufferedDraw() {
  return SendMsg(2034, 0, 0) != 0;
}

// If drawing is buffered then each line of text is drawn into a bitmap buffer
// before drawing it to the screen to avoid flicker.
void SetBufferedDraw(bool buffered) {
  SendMsg(2035, buffered, 0);
}



// Set the code page used to interpret the bytes of the document as characters.
// The SC_CP_UTF8 value can be used to enter Unicode mode.
void SetCodePage(FXint codePage) {
  SendMsg(2037, codePage, 0);
}



// Set the set of characters making up words for when moving or selecting
// by word.
void SetWordChars(const FXString& characters) {
  SendMsg(2077, 0, (long)characters.text());
}

// Show or hide the horizontal scroll bar.
void SetUseHorizontalScrollBar(bool show) {
  SendMsg(2130, show, 0);
}

// Is the horizontal scroll bar visible?
bool GetUseHorizontalScrollBar() {
  return SendMsg(2131, 0, 0) != 0;
}

// From WX
// On Windows will draw the document into a display context such as a printer.
/* int FormatRange( bool   doDraw,
                    int    startPos,
                    int    endPos,
                    wxDC*  draw,
                    wxDC*  target,
                    wxRect renderRect,
                    wxRect pageRect) {
  RangeToFormat fr;
  fr.hdc = draw;
  fr.hdcTarget = target;
  fr.rc.top = renderRect.GetTop();
  fr.rc.left = renderRect.GetLeft();
  fr.rc.right = renderRect.GetRight();
  fr.rc.bottom = renderRect.GetBottom();
  fr.rcPage.top = pageRect.GetTop();
  fr.rcPage.left = pageRect.GetLeft();
  fr.rcPage.right = pageRect.GetRight();
  fr.rcPage.bottom = pageRect.GetBottom();
  fr.chrg.cpMin = startPos;
  fr.chrg.cpMax = endPos;
  return SendMsg(2151, doDraw, (long)&fr);
}*/


// Set to read only or read write.
void SetReadOnly(bool readOnly) {
  SendMsg(2171, readOnly, 0);
}

// Set to overtype (true) or insert mode
void SetOvertype(bool overtype) {
  SendMsg(2186, overtype, 0);
}

// Returns true if overtype mode is active otherwise false is returned.
bool GetOvertype() {
  return SendMsg(2187, 0, 0) != 0;
}

// Set whether a pop up menu is displayed automatically when the user presses
// the wrong mouse button.
void UsePopUp(bool allowPopUp) {
  SendMsg(2371, allowPopUp, 0);
}

// Set the zoom level. This number of points is added to the size of all fonts.
// It may be positive to magnify or negative to reduce.
void SetZoom(int zoom) {
  SendMsg(2373, zoom, 0);
}

// Retrieve the zoom level.
int GetZoom() {
  return SendMsg(2374, 0, 0);
}


#endif




