// Scintilla source code edit control
// ScintillaFOX.cpp - implementation for the FOX tollkit
// The License.txt file describes the conditions under which this software may be distributed.
// SB: initial release; based on ScintillaWX by Robin Dunn.


#include <ctype.h>
#include <sys/time.h>
#include <unistd.h>
#include "ScintillaFOX.h"
#include "stc.h"




/*
class wxSTCCallTip : public wxWindow {
public:
    wxSTCCallTip(wxWindow* parent, int ID, CallTip* ct)
        : wxWindow(parent, ID)
        {
            m_ct = ct;
        }

    void OnPaint(wxPaintEvent& evt) {
        wxPaintDC dc(this);
        Surface surfaceWindow;
        surfaceWindow.Init(&dc);
        m_ct->PaintCT(&surfaceWindow);
        surfaceWindow.Release();
    }

    CallTip*    m_ct;
    DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(wxSTCCallTip, wxWindow)
    EVT_PAINT(wxSTCCallTip::OnPaint)
END_EVENT_TABLE()

*/


//----------------------------------------------------------------------
// Constructor/Destructor


ScintillaFX::ScintillaFX(FXStyledText* win) {
    wMain = win;
    stc = win;
    Initialise();
}


ScintillaFX::~ScintillaFX() {
    SetTicking(false);
}

//----------------------------------------------------------------------
// base class virtuals


void ScintillaFX::Initialise() {
}


void ScintillaFX::Finalise() {
  ScintillaBase::Finalise();
}


void ScintillaFX::StartDrag() {
//printf("ScintillaFox: StartDrag() called\n");
//SB TODO: add config option "Cut/Drag with or wo EOL" (def is with)
  stc->m_dragtext = dragChars;
	stc->flags |= stc->FLAG_TRYDRAG;
}


void ScintillaFX::SetTicking(bool on) {
//printf("ScintillaFOX: SetTicking(%d) called\n", on);
  if (on && !stc->blinker) stc->StartBlinker();
  else if (!on && stc->blinker) stc->StopBlinker();
}


void ScintillaFX::SetMouseCapture(bool on) {
//printf("->ScintillaFOX: SetMouseCapture(%d) called\n", on);
  if(on && !stc->grabbed()) stc->grab();
  else if (!on && stc->grabbed()) stc->ungrab();
}


bool ScintillaFX::HaveMouseCapture() {
//printf("->ScintillaFOX: HaveMouseCapture() called, returning %d\n", m_IsCaptured);
  return stc->grabbed();
}


////////// SCROLLING //////////

void ScintillaFX::ScrollText(int linesToMove) {
// SB: This function is called by Editor when using V-scrollbar (dragging or scrolling)
// to scroll the canvas. After this one, the Editor calls SetVerticalScrollPos()
// to adjust the scrollbar.
// This function shouldn't adjust the scrollbar, othervise it'll be adjusted second
// time in SetVerticalScrollPos(). Because moveContents() does adjust the pos_y, we'll
// filter out this event in SetVerticalScrollPos().

//printf("1. ScintillaFOX: ScrollText() topLine= %d, lines_tomove= %d\n",topLine, linesToMove);

  FXint new_y = -topLine * stc->m_itemHeight;
  stc->moveContents(stc->pos_x, new_y);
}


void ScintillaFX::SetVerticalScrollPos() {
// SB: This function is called by Editor when using keyboard to scroll the canvas.
// The scrolling is handled by Editor, here we just need to adjust the scrollbar.
// This function is *also* called after using the scrollbar (dragging or scrolling).
// The ScrollText() function should not adjust the scrollbar, but because is does
// (via moveContents) we return if we got here from ScrollText() (pos_y == new_y),
// but only if we are not autoscrolling (!grabbed()).

//printf("2. ScintillaFOX: SetVerticalScrollPos() topLine= %d\n", topLine);

  FXint new_y = -topLine * stc->m_itemHeight;
  if (stc->pos_y == new_y && !stc->grabbed()) {
    //printf ("ScintillaFX: SetVerticalScrollPos() skipping line %d!\n", topLine);
    return;
  }
  stc->vertical->setPosition(-new_y);
  stc->pos_y = new_y;
}


void ScintillaFX::SetHorizontalScrollPos() {
//printf("3. ScintillaFOX: SetHorizontalScrollPos() xOffset=%d\n", xOffset);

  stc->horizontal->setPosition(xOffset);
  stc->pos_x = -xOffset;
}


bool ScintillaFX::ModifyScrollBars(int nMax, int nPage) {
//printf("4. ScintillaFOX: ModifyScrollBars() nMax= %d, nPage= %d\n", nMax, nPage);

  // modify the vertical scrollbar; Todo: horizontal scrollbar
  stc->m_nitems = nMax;

  stc->layout();
  return 0;  //true = do Redraw()
}



void ScintillaFX::NotifyChange() {
  stc->handle( FXSciObject(wMain.GetID()), MKUINT(ID_SCINTILLA, SCIMSG_CHANGE), 0);
}


void ScintillaFX::NotifyParent(SCNotification scn) {
  stc->handle( FXSciObject(wMain.GetID()), MKUINT(ID_SCINTILLA, SCIMSG_NOTIFY), &scn);
}


//// Clipboard

void ScintillaFX::Copy() {
  FXDragType types[1];

  if (currentPos != anchor) {
    char* text = CopySelectionRange();
    types[0] = stc->stringType;

    if(stc->acquireClipboard(types, 1)) {
      FXFREE(&stc->clipbuffer);
      stc->cliplength = SelectionRangeLength();
      if (stc->cliplength <= 0) return;

      FXCALLOC(&stc->clipbuffer, FXchar, stc->cliplength +1);
      if (!stc->clipbuffer) return;

      strcpy(stc->clipbuffer, text);
    }
  }
}


void ScintillaFX::Paste() {
  pdoc->BeginUndoAction();
  ClearSelection();
  FXuchar *data; FXuint len;

  if(stc->getDNDData(FROM_CLIPBOARD, stc->stringType, data, len)) {
    FXRESIZE(&data, FXuchar, len+1); data[len]='\0';
#ifdef WIN32
    FXuint i,j;
    FXuchar c;
    for(i=j=0; j<len; j++){if((c=data[j])!='\r'){data[i++]=c;}}
    data[i]='\0';
#endif
    pdoc->InsertString(currentPos, (char*)data, (int)len);
    SetEmptySelection(currentPos + len);
    FXFREE(&data);
  }

  pdoc->EndUndoAction();
  NotifyChange();
  Redraw();
}


bool ScintillaFX::CanPaste() {
  FXuchar *data; FXuint len;
  return (stc->getDNDData(FROM_CLIPBOARD,stc->stringType,data,len));
}




void ScintillaFX::CreateCallTipWindow(PRectangle) {
    //ct.wCallTip = new wxSTCCallTip(wDraw.GetID(), -1, &ct);
    //ct.wDraw = ct.wCallTip;
}


void ScintillaFX::AddToPopUp(const char *label, int cmd, bool enabled) {
  FXMenuCommand* menu;

  if (!label[0]) new FXMenuSeparator(stc->m_pane);
  else menu = new FXMenuCommand(stc->m_pane, label, NULL, stc, cmd);

  switch(cmd) {
    case idCallTip:       menu->setSelector(stc->ID_CALLTIP);       break;
    case idAutoComplete:  menu->setSelector(stc->ID_AUTOCOMPLETE);  break;
    case idcmdUndo:       menu->setSelector(stc->ID_UNDO);          break;
    case idcmdRedo:       menu->setSelector(stc->ID_REDO);          break;
    case idcmdCut:        menu->setSelector(stc->ID_CUT_SEL);       break;
    case idcmdCopy:       menu->setSelector(stc->ID_COPY_SEL);      break;
    case idcmdPaste:      menu->setSelector(stc->ID_PASTE_SEL);     break;
    case idcmdDelete:     menu->setSelector(stc->ID_DELETE_SEL);    break;
    case idcmdSelectAll:  menu->setSelector(stc->ID_SELECTALL);     break;
  }

  //if (!enabled)
  FXuint msg = enabled ? FXWindow::ID_ENABLE : FXWindow::ID_DISABLE;
  menu->handle( FXSciObject(wMain.GetID()), MKUINT(msg, SEL_COMMAND), NULL);

}


void ScintillaFX::ClaimSelection() {
  stc->grabSelection();
}


// SB: (iMessage, wParam, lParam)
long ScintillaFX::DefWndProc(unsigned int, unsigned long, long) {
  return 0;
}

long ScintillaFX::WndProc(unsigned int iMessage, unsigned long wParam, long lParam) {
  return ScintillaBase::WndProc(iMessage, wParam, lParam);
}



//----------------------------------------------------------------------
// Event delegates

void ScintillaFX::DoPaint(FXDrawable* dc, FXRectangle rect) {
//printf("ScintillaFOX: DoPaint() repaint_rect= %d, %d, %d, %d\n",
//       rect.x, rect.y, rect.w, rect.h);

  paintState = painting;
  Surface surfaceWindow;
  surfaceWindow.Init(dc);
  //WARNING: we are filling in the Editor::rcPaint variable !
  rcPaint = PRectangleFromFXRectangle(rect);

  Paint(&surfaceWindow, rcPaint);

  surfaceWindow.Release();
  if (paintState == paintAbandoned) {
    // Painting area was insufficient
    FullPaint();
  }
  paintState = notPainting;
}



void ScintillaFX::DoHScroll(FXint x) {
  HorizontalScrollTo(x);
}

void ScintillaFX::DoVScroll(FXint y) {
  if (y != topLine) ScrollTo(y);
}


void ScintillaFX::DoSize() {
//printf("...xOffset=%d, stc->pos_x=%d\n", xOffset, stc->pos_x);

  // Y-scroll  // sets nMax, nPage, topLine; calls SetVerticalScrollPos()
  SetScrollBars();
  // X-scroll
  HorizontalScrollTo(-stc->pos_x);

  DropGraphics();
}



void ScintillaFX::DoLoseFocus(){
//printf("ScintillaFOX: DoLoseFocus() entered\n");
  SetFocusState(0);
}

void ScintillaFX::DoGainFocus(){
//printf("ScintillaFOX: DoGainFocus() entered\n");
  SetFocusState(1);
}

void ScintillaFX::DoSysColourChange() {
  InvalidateStyleData();
}


void ScintillaFX::DoAddChar(char ch) {
  AddChar(ch);
}


int  ScintillaFX::DoKeyDown(int key, bool shift, bool ctrl, bool alt) {
// printf("ScintillaFOX: DoKeyDown() key_code= %d\n", key);

  switch (key) {
    case KEY_Down:                      key = SCK_DOWN; break;
    case KEY_Up:                        key = SCK_UP; break;
    case KEY_Left:                      key = SCK_LEFT; break;
    case KEY_Right:                     key = SCK_RIGHT; break;
    case KEY_Home:                      key = SCK_HOME; break;
    case KEY_End:                       key = SCK_END; break;
    case KEY_Prior:                     key = SCK_PRIOR; break;
    case KEY_Next:                      key = SCK_NEXT; break;
    case KEY_Delete:                    key = SCK_DELETE; break;
    case KEY_Insert:                    key = SCK_INSERT; break;
    case KEY_Escape:                    key = SCK_ESCAPE; break;
    case KEY_BackSpace:                 key = SCK_BACK; break;
    case KEY_Tab:                       key = SCK_TAB; break;
    case KEY_Return:                    key = SCK_RETURN; break;
    case KEY_KP_Add:                    key = SCK_ADD; break;
    case KEY_KP_Subtract:               key = SCK_SUBTRACT; break;
    case KEY_KP_Divide:                 key = SCK_DIVIDE; break;
    case KEY_Control_L:                 key = 0; break;
    case KEY_Control_R:                 key = 0; break;
    case KEY_Alt_L:                     key = 0; break;
    case KEY_Alt_R:                     key = 0; break;
    case KEY_Shift_L:                   key = 0; break;
    case KEY_Shift_R:                   key = 0; break;
  }

  return KeyDown(key, shift, ctrl, alt);
}


void ScintillaFX::DoCommand(int ID) {
  Command(ID);  //not used
}


void ScintillaFX::DoContextMenu(Point pt) {
  ContextMenu(pt);
}

void ScintillaFX::DoOnListBox() {
  AutoCompleteCompleted();  //not used
}


//// Mouse

void ScintillaFX::DoButtonDown(Point pt, unsigned int curTime, bool shift, bool ctrl, bool alt) {
//printf("ScintillaFX: DoButtonDown() curTime= %u\n", curTime);
  if (PointInSelMargin(pt))
    stc->setDragCursor(stc->getApp()->getDefaultCursor(DEF_ARROW_CURSOR));

  ButtonDown(pt, curTime, shift, ctrl, alt);
}


void ScintillaFX::DoButtonUp(Point pt, unsigned int curTime, bool ctrl, bool moved) {
//printf("ScintillaFX: DoButtonUp() curTime= %u\n", curTime);

  //FXbool hack = inDragDrop;
  //inDragDrop = false;
  //ButtonUp(pt, curTime, ctrl);
  //inDragDrop = hack;


  if (HaveMouseCapture()) {
    xEndSelect = pt.x - vs.fixedColumnWidth + xOffset;
    ptMouseLast = pt;
    SetMouseCapture(false);
    lastClickTime = curTime;
    lastClick = pt;
    lastXChosen = pt.x;
    inDragDrop = false;
    EnsureCaretVisible(false);
  }

  if (PointInSelection(pt) &&  !moved && ((curTime - lastClickTime) != 0)){
    SetDragPosition(invalidPosition);
    MovePositionTo(PositionFromLocation(pt), false);
  }
  else if (inDragDrop)
    MovePositionTo(SelectionStart() + lenDrag, true);
}


void ScintillaFX::DoButtonMove(Point pt) {
//printf("ScintillaFX: DoButtonMove() called pt= (%u, %u)\n", pt.x, pt.y);
// SB Note: ButtonMove() *will* be executed on every blink from Editor::Tick()
// when HaveMouseCapture()==1, yielding autoscroll.

  ButtonMove(pt);
}


//----------------------------------------------------------------------

void ScintillaFX::DoDNDDrop(Point pt, FXString& text) {
//printf("DoDNDDrop called\n");
//TODO: find if rectangular...
  FXbool rectangular = false;
  FXint pos = PositionFromLocation(pt);
  // [pos, text, moving, rectangular]
  DropAt(pos, text.text(), false, rectangular);
  SetDragPosition(invalidPosition);
  //remove def. selection - interfere with X selection
  SetSelection(pos, CurrentPosition());
}

//----------------------------------------------------------------------


// Redraw all of text area. This paint will not be abandoned.
void ScintillaFX::FullPaint() {
//printf("ScintillaFox: FullPaint() called!!\n");
    paintState = painting;
    stc->update();
    paintState = notPainting;
}


void ScintillaFX::DoScrollToLine(int line) {
    ScrollTo(line);
}

void ScintillaFX::DoScrollToColumn(int column) {
    HorizontalScrollTo(column * vs.spaceWidth);
}


void ScintillaFX::DoTick() {
// SB: we could use 2 timers -> DoBlink and DoTick
// one for caret and one for autoscroll ...
//printf("DoTick()\n");
  Tick();
}



//SB
void ScintillaFX::SetCaretMargin(FXint chars) {
// Num. visible pixels on each side of caret
  if (chars < 2) chars = 2;
  xCaretMargin = chars * vs.spaceWidth;
}

FXbool ScintillaFX::IsPointInSelection(FXPoint& pt) {
  return PointInSelection(PointFromFXPoint(pt));
}

FXbool ScintillaFX::IsPointInSelMargin(FXPoint& pt) {
  return PointInSelMargin(PointFromFXPoint(pt));
}

FXint ScintillaFX::GetLineExtent(FXint line) {
// To be used for setting stc->content_w
//int line = the longest line
  RefreshStyleData();
	Surface surface;
	surface.Init();
	surface.SetUnicodeMode(SC_CP_UTF8 == pdoc->dbcsCodePage);
	LineLayout ll;
	LayoutLine(line, &surface, vs, ll);
	return (ll.positions[ll.numCharsInLine] + vs.fixedColumnWidth);
}

















