// Scintilla source code edit control
// ScintillaFOX.h - implementation for the FOX toolkit
// The License.txt file describes the conditions under which this software may be distributed.
// SB: initial release; based on ScintillaWX by Robin Dunn.

#ifndef __ScintillaFX_h__
#define __ScintillaFX_h__

#include "Platform.h"

#include "Scintilla.h"
//#ifdef SCI_LEXER
#include "SciLexer.h"
#include "PropSet.h"
#include "Accessor.h"
#include "KeyWords.h"
//#endif

#include "ContractionState.h"
#include "SVector.h"
#include "CellBuffer.h"
#include "CallTip.h"
#include "KeyMap.h"
#include "Indicator.h"
#include "LineMarker.h"
#include "Style.h"
#include "ViewStyle.h"
#include "AutoComplete.h"
#include "Document.h"
#include "Editor.h"
#include "ScintillaBase.h"

#include <fox/fx.h>
#include <fox/fxkeys.h>

//SB:
//
// WX scheme:
//    wxStyledTextCtrl  -->  ScintillaWX  -->  Editor [Scintilla proper]
//    OnChar()          -->  DoAddChar()  -->  AddChar()
//
// FOX preliminary:
//    FXStyledText      -->  ScintillaFX  -->  Editor [Scintila proper]
//    OnCmdInsertChar() -->  DoAddChar()  -->  AddChar()


#define FXSciObject(a) ((FXObject* ) a)


class FXStyledText;
class ScintillaFX;


class ScintillaFX : public ScintillaBase {
public:
  ScintillaFX(FXStyledText* win);
  ~ScintillaFX();

  // Event delegates
  void DoPaint(FXDrawable* dc, FXRectangle rect);
  void DoHScroll(FXint pos);
  void DoVScroll(FXint pos);
  void DoSize();
  void DoLoseFocus();
  void DoGainFocus();
  void DoSysColourChange();
  void DoButtonDown(Point pt, unsigned int curTime, bool shift, bool ctrl, bool alt);
  void DoButtonUp(Point pt, unsigned int curTime, bool ctrl, bool moved);
  void DoButtonMove(Point pt);
  void DoAddChar(char ch);
  int  DoKeyDown(int key, bool shift, bool ctrl, bool alt);
  void DoTick();
  void DoCommand(int ID);
  void DoContextMenu(Point pt);
  void DoOnListBox();

  // helpers
  void FullPaint();
  bool CanPaste();
  bool GetHideSelection() { return hideSelection; }
  void DoScrollToLine(int line);
  void DoScrollToColumn(int column);

  // base class virtuals
  virtual void Initialise();
  virtual void Finalise();
  virtual void StartDrag();
  virtual void SetTicking(bool on);
  virtual void SetMouseCapture(bool on);
  virtual bool HaveMouseCapture();
  virtual void ScrollText(int linesToMove);
  virtual void SetVerticalScrollPos();
  virtual void SetHorizontalScrollPos();
  virtual bool ModifyScrollBars(int nMax, int nPage);
  virtual void Copy();
  virtual void Paste();
  virtual void CreateCallTipWindow(PRectangle rc);
  virtual void AddToPopUp(const char *label, int cmd = 0, bool enabled = true);
  virtual void ClaimSelection();

  virtual long DefWndProc(unsigned int iMessage,
                          unsigned long wParam,
                          long lParam);

  virtual long WndProc(unsigned int iMessage,
                       unsigned long wParam,
                       long lParam);

  virtual void NotifyChange();
  virtual void NotifyParent(SCNotification scn);


//SB
public:
  void SetCaretMargin(FXint chars);
  void DoDNDDrop(Point pt, FXString& text);
  FXbool IsPointInSelection(FXPoint& fxpt);
  FXbool IsPointInSelMargin(FXPoint& fxpt);
  FXint GetLineExtent(FXint line);

  enum {
    ID_SCINTILLA     = FXWindow::ID_LAST,
    SCIMSG_CHANGE,
    SCIMSG_NOTIFY,

    ID_LAST,
  };


private:
  FXStyledText* stc;

};



#endif
