// seditor.cpp - Test editor using the Scintilla cource code edit control
// implemented for the FOX toolkit
// SB: initial release;

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <signal.h>
#include <fox/fx.h>
#include "seditor.h"
#include "Scintilla.h"


// Map
FXDEFMAP(StyledTextWindow) StyledTextWindowMap[]={
  FXMAPFUNC(SEL_COMMAND,     StyledTextWindow::ID_OPEN,   StyledTextWindow::onCmdOpen),
  FXMAPFUNC(SEL_COMMAND,     StyledTextWindow::ID_QUIT,   StyledTextWindow::onCmdExit),
  FXMAPFUNC(SEL_COMMAND,     StyledTextWindow::ID_ABOUT,  StyledTextWindow::onCmdHelp),
	FXMAPFUNC(SEL_DND_MOTION,  StyledTextWindow::ID_TEXT,   StyledTextWindow::onEditDNDMotion),
	FXMAPFUNC(SEL_DND_DROP,    StyledTextWindow::ID_TEXT,   StyledTextWindow::onEditDNDDrop),
};


// Object implementation
FXIMPLEMENT(StyledTextWindow, FXMainWindow, StyledTextWindowMap, ARRAYNUMBER(StyledTextWindowMap))


// Make some windows
StyledTextWindow::StyledTextWindow(FXApp* a) :
                  FXMainWindow(a,"Styled Text Editor: - untitled",
                  NULL, NULL, DECOR_ALL, 0,0, 800,600,0,0){


  // Make main window; set myself as the target
  setTarget(this);
  setSelector(ID_TITLE);


  // Make some icons
  newicon   = new FXGIFIcon(getApp(), newdoc);
  openicon  = new FXGIFIcon(getApp(), opendoc);
  saveicon  = new FXGIFIcon(getApp(), savedoc);
  cuticon   = new FXGIFIcon(getApp(), cutdoc);
  copyicon  = new FXGIFIcon(getApp(), copydoc);
  pasteicon = new FXGIFIcon(getApp(), pastedoc);
  printicon = new FXGIFIcon(getApp(), printdoc);


  // Menu bar
  FXToolbarShell* dragshell1 = new FXToolbarShell(
                  this,
                  FRAME_RAISED|FRAME_THICK,
                  0,0, 0,0,0,0);

  FXMenubar* menubar = new FXMenubar(
             this,
             dragshell1,
             LAYOUT_SIDE_TOP | LAYOUT_FILL_X,
             0,0,0,0, 3,3,10,10, 10,10);

  new FXToolbarGrip(menubar,menubar,FXMenubar::ID_TOOLBARGRIP,TOOLBARGRIP_DOUBLE);


  // Separator
  new FXHorizontalSeparator(this);


  // Tool bar
  FXToolbarShell* dragshell2 = new FXToolbarShell(
                  this,
                  FRAME_RAISED|FRAME_THICK,
                  0,0, 0,0,0,0);

  FXToolbar* toolbar = new FXToolbar(
             this,
             dragshell2,
             LAYOUT_SIDE_TOP | LAYOUT_FILL_X | PACK_UNIFORM_WIDTH | PACK_UNIFORM_HEIGHT,
             0,0,0,0, 3,3,2,2, 0,0);

  new FXToolbarGrip(toolbar,toolbar,FXToolbar::ID_TOOLBARGRIP,TOOLBARGRIP_DOUBLE);


  // Status bar
  FXStatusbar* statusbar = new FXStatusbar(
               this,
               LAYOUT_SIDE_BOTTOM | LAYOUT_FILL_X,
               0,0,0,0, 0,0,0,0);


  // Sunken border for text widget
  FXHorizontalFrame* textbox = new FXHorizontalFrame(
                     this,
                     FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X|LAYOUT_FILL_Y,
                     0,0,0,0, 0,0,0,0);


  // Make editor window
  editor=new FXStyledText(
             textbox,
             this,
             ID_TEXT,
             LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_RAISED);

  editor->setFocus();


  // Menu bar entries

  // File menu
  filemenu=new FXMenuPane(this);
  new FXMenuTitle(menubar,"&File",NULL,filemenu);

  // Edit Menu
  editmenu=new FXMenuPane(this);
  new FXMenuTitle(menubar,"&Edit",NULL,editmenu);

  // Goto Menu
  gotomenu=new FXMenuPane(this);
  new FXMenuTitle(menubar,"&Goto",NULL,gotomenu);

  // Search Menu
  searchmenu=new FXMenuPane(this);
  new FXMenuTitle(menubar,"&Search",NULL,searchmenu);

  // Options Menu
  optionmenu=new FXMenuPane(this);
  new FXMenuTitle(menubar,"&Options",NULL,optionmenu);

  // View menu
  viewmenu=new FXMenuPane(this);
  new FXMenuTitle(menubar,"&View",NULL,viewmenu);

  // Help menu
  helpmenu=new FXMenuPane(this);
  new FXMenuTitle(menubar,"&Help",NULL,helpmenu,LAYOUT_RIGHT);


  // File Menu entries
  new FXMenuCommand(filemenu,"&Open...   \tCtl-O\tOpen document file.",NULL,this,ID_OPEN);
  new FXMenuCommand(filemenu,"&New...\tCtl-N\tCreate new document.",NULL,this,ID_NEW);
  new FXMenuCommand(filemenu,"&Save\tCtl-S\tSave changes to file.",NULL,this,ID_SAVE);
  new FXMenuCommand(filemenu,"Save &As...\t\tSave document to another file.",NULL,this,ID_SAVEAS);
  new FXMenuSeparator(filemenu);
  new FXMenuCommand(filemenu,"&Print...\tCtl-P\tPrint document.",NULL,this,ID_PRINT);
  new FXMenuCommand(filemenu,"&Exit\tCtl-X\tExit.",NULL,this,ID_QUIT);

 // Edit Menu entries
  new FXMenuCommand(editmenu,"&Undo   \tCtl-Z\tUndo last change.",NULL,editor,FXStyledText::ID_UNDO);
  new FXMenuCommand(editmenu,"&Redo\tCtl-Y\tRedo last undo.",NULL,editor,FXStyledText::ID_REDO);
  new FXMenuSeparator(editmenu);
  new FXMenuCommand(editmenu,"&Copy\tCtl-C\tCopy selection to clipboard.",NULL,editor,FXStyledText::ID_COPY_SEL);
  new FXMenuCommand(editmenu,"Cu&t\tCtl-X\tCut selection to clipboard.",NULL,editor,FXStyledText::ID_CUT_SEL);
  new FXMenuCommand(editmenu,"&Paste\tCtl-V\tPaste from clipboard.",NULL,editor,FXStyledText::ID_PASTE_SEL);
  new FXMenuCommand(editmenu,"&Delete\t\tDelete selection.",NULL,editor,FXStyledText::ID_DELETE_SEL);

  // Goto Menu entries
  new FXMenuCommand(gotomenu,"&Goto...   \tCtl-G\tGoto line number.",NULL,editor,FXStyledText::ID_GOTO_LINE);

  // Search Menu entries
  new FXMenuCommand(searchmenu,"&Search...\tCtl-F\tSearch for a string.",NULL,editor,FXStyledText::ID_SEARCH);
  new FXMenuCommand(searchmenu,"R&eplace...   \tCtl-R\tSearch for a string.",NULL,editor,FXStyledText::ID_REPLACE);

  // Options menu
  new FXMenuCommand(optionmenu,"Preferences...\t\tChange preferences.",NULL,this,ID_PREFERENCES);
  new FXMenuCommand(optionmenu,"Font...\t\tChange text font.",NULL,this,ID_FONT);
  new FXMenuCommand(optionmenu,"Save Settings...\t\tSave settings now.",NULL,this,ID_SAVE_SETTINGS);

  // View Menu entries
  new FXMenuCommand(viewmenu,"Toolbar      \t \tDisplay toolbar.",NULL,toolbar,FXWindow::ID_TOGGLESHOWN);
  new FXMenuCommand(viewmenu,"Status line\t\tDisplay status line.",NULL,statusbar,FXWindow::ID_TOGGLESHOWN);
  new FXMenuSeparator(viewmenu);
	new FXMenuCommand(viewmenu,"Line numbers\t\tDisplay line numbers.",NULL,editor,FXStyledText::ID_LINENUM_MARGIN);
	new FXMenuCommand(viewmenu,"Symbol margin\t\tDisplay symbol margin.",NULL,editor,FXStyledText::ID_SYMBOL_MARGIN);
	new FXMenuCommand(viewmenu,"Fold margin\t\tDisplay fold margin.",NULL,editor,FXStyledText::ID_FOLD_MARGIN);

  //Help Menu entries
  new FXMenuCommand(helpmenu,"&Help...   \t \tDisplay help information.",NULL,this,ID_SHOW_HELP,0);
  new FXMenuSeparator(helpmenu);
  new FXMenuCommand(helpmenu,"&About seditor...\t\tDisplay about panel.",NULL,this,ID_ABOUT,0);


  // Tool bar buttons

  // New
  butNew = new FXButton(
            toolbar,
            "\tNew\tNew document file.",
            newicon,
            this,
            ID_NEW,
            ICON_ABOVE_TEXT | BUTTON_TOOLBAR | FRAME_RAISED | LAYOUT_TOP | LAYOUT_LEFT,
            0,0,0,0, 5,5,5,5);

  // Open
  butOpen = new FXButton(
            toolbar,
            "\tOpen\tOpen document file.",
            openicon,
            this,
            ID_OPEN,
            ICON_ABOVE_TEXT | BUTTON_TOOLBAR | FRAME_RAISED | LAYOUT_TOP | LAYOUT_LEFT,
            0,0,0,0, 5,5,5,5);

  // Save
  butSave = new FXButton(
            toolbar,
            "\tSave\tSave document file.",
            saveicon,
            this,
            ID_SAVE,
            ICON_ABOVE_TEXT | BUTTON_TOOLBAR | FRAME_RAISED | LAYOUT_TOP | LAYOUT_LEFT,
            0,0,0,0, 5,5,5,5);

  // Cut
  butCut = new FXButton(
            toolbar,
            "\tCut\tCut the selection.",
            cuticon,
            editor,
            FXStyledText::ID_CUT_SEL,
            ICON_ABOVE_TEXT | BUTTON_TOOLBAR | FRAME_RAISED | LAYOUT_TOP | LAYOUT_LEFT,
            0,0,0,0, 5,5,5,5);

  // Copy
  butCopy = new FXButton(
            toolbar,
            "\tCopy\tCopy the selection.",
            copyicon,
            editor,
            FXStyledText::ID_COPY_SEL,
            ICON_ABOVE_TEXT | BUTTON_TOOLBAR | FRAME_RAISED | LAYOUT_TOP | LAYOUT_LEFT,
            0,0,0,0, 5,5,5,5);

  // Paste
  butPaste = new FXButton(
            toolbar,
            "\tPaste\tPaste from clipboard.",
            pasteicon,
            editor,
            FXStyledText::ID_PASTE_SEL,
            ICON_ABOVE_TEXT | BUTTON_TOOLBAR | FRAME_RAISED | LAYOUT_TOP | LAYOUT_LEFT,
            0,0,0,0, 5,5,5,5);

  // Print
  butPrint = new FXButton(
            toolbar,
            "\tPrint\tPrint document file.",
            printicon,
            this,
            ID_PRINT,
            ICON_ABOVE_TEXT | BUTTON_TOOLBAR | FRAME_RAISED | LAYOUT_TOP | LAYOUT_LEFT,
            0,0,0,0, 5,5,5,5);


  new FXVerticalSeparator(toolbar);


  // Make a tool tip
  new FXTooltip(getApp(),0);

	// Init variables:
  m_filename = "untitled";
  m_stripcr  = TRUE;
  m_stripsp  = FALSE;

}


// Clean up the mess
StyledTextWindow::~StyledTextWindow(){
  delete filemenu;
  delete editmenu;
  delete gotomenu;
  delete searchmenu;
  delete optionmenu;
  delete viewmenu;
  delete helpmenu;
  delete newicon;
  delete openicon;
}


//////////////////////////////////////////////////////////////////////////////
//// Messages:


// Open (from TextEdit.cpp)
long StyledTextWindow::onCmdOpen(FXObject*,FXSelector,void*){
  FXString file;
  //if(!saveChanges()) return 1;
  FXFileDialog opendialog(this,"Open Document");

  opendialog.setSelectMode(SELECTFILE_EXISTING);
  //opendialog.setPatternList(getPatterns());
  //opendialog.setCurrentPattern(getCurrentPattern());
  opendialog.setFilename(m_filename);

  if(opendialog.execute()){
    //setCurrentPattern(opendialog.getCurrentPattern());
    file=opendialog.getFilename();
    LoadFile(file);
  }

	return 1;
}


// Exit
long StyledTextWindow::onCmdExit(FXObject*,FXSelector,void*) {
  getApp()->exit(0);
  return 0;
}


// Help
long StyledTextWindow::onCmdHelp(FXObject*,FXSelector,void*) {
  return 0;
}


// DND motion
long StyledTextWindow::onEditDNDMotion(FXObject*,FXSelector,void*) {
//printf("seditor: onEditDNDMotion() called\n");

  if(offeredDNDType(FROM_DRAGNDROP,urilistType)){
    acceptDrop(DRAG_COPY);
    return 1;
  }

  return 0;
}


// DND drop
long StyledTextWindow::onEditDNDDrop(FXObject*,FXSelector,void*) {
//printf("seditor: onEditDNDDrop() called\n");

  FXchar *data; FXuint len;

  if(getDNDData(FROM_DRAGNDROP,urilistType,(FXuchar*&)data,len)){
    FXString urilist(data,len);
    FXString file=FXURL::fileFromURL(urilist.before('\r'));
    FXFREE(&data);

    if(file.empty()) return 1;
    //if(!saveChanges()) return 1;

		printf("seditor: loading %s\n", file.text());
    LoadFile(file);
    return 1;
  }

	return 0;
}


//////////////////////////////////////////////////////////////////////////////
//// Functions:

// Set font
void StyledTextWindow::SetFont(const FXString& face, int size) {
  editor->SetFont(face, size);    //<--
}


// Load file (from TextEdit.cpp)
FXbool StyledTextWindow::LoadFile(const FXString& file){
//SB: TODO: filter non-acceptable files (binaries, dirs)

  unsigned long size,n,i,j;
  char *text;
  FILE *fp;

  FXTRACE((100,"loadFile(%s)\n",file.text()));

  // Open file
  fp=fopen(file.text(),"r");
  if(!fp){
    FXMessageBox::error(this,MBOX_OK,"Error Loading File","Unable to open file: %s",file.text());
    return FALSE;
  }

  // Get file size
  size=FXFile::size(file);

  // Make buffer to load file
  if(!FXCALLOC(&text,FXchar,size+1)){
    FXMessageBox::error(this,MBOX_OK,"Error Loading File","File is too big: %s",file.text());
    fclose(fp);
    return FALSE;
  }

  // Set wait cursor
  getApp()->beginWaitCursor();

  // Read the file
  n=fread(text,1,size,fp);

  // Strip carriage returns
  if(m_stripcr){
    for(i=j=0; j<n; j++){
      if(text[j]!='\r') text[i++]=text[j];
    }
    text[i]='\0';
    n=i;
  }

  // Set text
  editor->SetText(text);    //<--
  FXFREE(&text);

  // Kill wait cursor
  getApp()->endWaitCursor();

  // Close file
  fclose(fp);

  // Set stuff
  //editor->setModified(FALSE);
  //editor->setEditable(FXFile::isWritable(file));
  //dirlist->setCurrentFile(file);
  //mrufiles.appendFile(file);
  //filetime=FXFile::modified(file);
  //filename=file;
  //filenameset=TRUE;
  //undolist.clear();
  //undolist.mark();

  return TRUE;
}


//////////////////////////////////////////////////////////////////////////////


// Start the ball rolling
void StyledTextWindow::start(int& argc,char** argv){
  if(argc>1){
    LoadFile(FXFile::absolute(argv[1]));
  }
}


// Create and show window
void StyledTextWindow::create(){
  if(!urilistType){urilistType=getApp()->registerDragType(urilistTypeName);}
  //readRegistry();
  FXMainWindow::create();
  show(PLACEMENT_SCREEN);
  //timer=getApp()->addTimeout(1000,this,ID_FILETIME);
}


//////////////////////////////////////////////////////////////////////////////


// Start the whole thing
int main(int argc,char *argv[]){
//fxTraceLevel = 151;

  // Make application
  FXApp application("StyledTextEdit","FoxStyledTest");

  // Open display
  application.init(argc,argv);

  // Make window
  StyledTextWindow* window=new StyledTextWindow(&application);

  // Handle interrupt to save stuff nicely
  application.addSignal(SIGINT,window,StyledTextWindow::ID_QUIT);

  // Create it
  application.create();

  // Start
  window->start(argc,argv);

  // Run
  return application.run();
}





