// seditor.h - Test editor using the Scintilla cource code edit control
// implemented for the FOX toolkit
// SB: initial release;

#ifndef SEDITOR_H
#define SEDITOR_H

#include "stc.h"


class FXAPI StyledTextWindow : public FXMainWindow {
  FXDECLARE(StyledTextWindow)

protected:
  FXHorizontalFrame*  textbox;
  FXStyledText*       editor;               // The StyledText widget
  FXMenuPane          *filemenu;            // File menu
  FXMenuPane          *editmenu;            // Edit menu
  FXMenuPane          *gotomenu;            // Goto menu
  FXMenuPane          *searchmenu;          // Search menu
  FXMenuPane          *optionmenu;          // Option menu
  FXMenuPane          *viewmenu;            // View menu
  FXMenuPane          *helpmenu;            // Help menu
  FXIcon              *newicon;
  FXIcon              *openicon;
  FXIcon              *saveicon;
  FXIcon              *cuticon;
  FXIcon              *copyicon;
  FXIcon              *pasteicon;
  FXIcon              *printicon;

  FXButton*           butNew;
  FXButton*           butOpen;
  FXButton*           butSave;
  FXButton*           butCut;
  FXButton*           butCopy;
  FXButton*           butPaste;
  FXButton*           butPrint;
	FXString            m_filename;           // File being edited
	FXbool              m_stripcr;            // Strip carriage returns
	FXbool              m_stripsp;            // Strip trailing spaces


protected:
  StyledTextWindow(){};


public:
  long onCmdOpen       (FXObject*, FXSelector, void*);
  long onCmdExit       (FXObject*, FXSelector, void*);
  long onCmdHelp       (FXObject*, FXSelector, void*);
  long onEditDNDMotion (FXObject*, FXSelector, void*);
	long onEditDNDDrop   (FXObject*, FXSelector, void*);


public:
  enum {
  ID_TITLE = FXMainWindow::ID_LAST,
  ID_NEW,
  ID_OPEN,
  ID_SAVE,
  ID_SAVEAS,
  ID_PRINT,
  ID_TEXT,
  ID_QUIT,
  ID_PREFERENCES,
  ID_FONT,
  ID_SAVE_SETTINGS,
  ID_SHOW_HELP,
  ID_ABOUT,

  ID_LAST
};


public:
  StyledTextWindow(FXApp* a);
  virtual void create();
  void start(int& argc, char* argv[]);
  virtual ~StyledTextWindow();

//SB:
  void SetFont(const FXString& face, int size);
	FXbool LoadFile(const FXString& file);

};


// Icons:

/* Generated by reswrap from file new.gif */
const unsigned char newdoc[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x0f,0x00,0xa1,0x00,0x00,0xc0,0xc0,0xc0,
  0x00,0x00,0x00,0xff,0xff,0xff,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,
  0x0f,0x00,0x00,0x02,0x2a,0x84,0x83,0x89,0xc6,0x1c,0x02,0xe3,0x6a,0x27,0x4a,0x11,
  0x28,0x78,0x16,0x27,0xca,0x75,0x50,0xd6,0x84,0x22,0xe9,0x88,0x12,0xa8,0x8e,0x6c,
  0x8b,0x56,0xf0,0xab,0xc6,0x5b,0x8b,0x81,0xca,0x6e,0x6b,0xfe,0xff,0x2b,0x00,0x00,
  0x3b
};

/* Generated by reswrap from file open.gif */
const unsigned char opendoc[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x0f,0x00,0xc2,0x00,0x00,0xc0,0xc0,0xc0,
  0x00,0x00,0x00,0xff,0xff,0xff,0xff,0xff,0x00,0x80,0x80,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,0x0f,0x00,0x00,0x03,
  0x35,0x08,0xba,0x1c,0xce,0x70,0x05,0x09,0x66,0xbc,0xaf,0xe6,0xe8,0x66,0x10,0x83,
  0xd0,0x8d,0x50,0x10,0x9e,0x20,0x68,0x49,0x29,0x1a,0xae,0x8a,0x99,0x8e,0xb4,0x37,
  0x13,0x78,0x9e,0x7b,0xaf,0xee,0x5b,0x9f,0x80,0x4f,0xb7,0x72,0x0c,0x77,0x8d,0x1a,
  0xe9,0xc2,0x6c,0x3a,0x01,0x09,0x00,0x3b
};

/* Generated by reswrap from file save.gif */
const unsigned char savedoc[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x0f,0x00,0xa1,0x00,0x00,0xc0,0xc0,0xc0,
  0x00,0x00,0x00,0x80,0x80,0x00,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,
  0x0f,0x00,0x00,0x02,0x31,0x44,0x8e,0x99,0x60,0x12,0x0c,0xe1,0x7b,0x8d,0x46,0x86,
  0xb0,0xbb,0x72,0x57,0xae,0x51,0x81,0xc7,0x8d,0x22,0x79,0x99,0x9a,0xa3,0x1c,0x82,
  0x37,0xbe,0xf2,0xfc,0x9e,0x6c,0xab,0x36,0x77,0xf6,0xe9,0x6d,0xb8,0xfa,0xf5,0x0c,
  0x38,0x05,0xe8,0xc8,0x28,0x00,0x00,0x3b
};

/* Generated by reswrap from file cut.gif */
const unsigned char cutdoc[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x0f,0x00,0xa1,0x00,0x00,0xc0,0xc0,0xc0,
  0x00,0x00,0x00,0x00,0x00,0x80,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,
  0x0f,0x00,0x00,0x02,0x2a,0x84,0x1f,0x78,0xbb,0x69,0xce,0x9a,0x8a,0x2d,0x05,0x48,
  0x1f,0xc0,0x79,0xdf,0xce,0x70,0x9d,0x10,0x08,0xe0,0x61,0x0a,0xea,0xb8,0x9a,0x80,
  0xcb,0xb8,0xa9,0x01,0x2f,0xf2,0x4b,0x53,0xf7,0x5b,0xc7,0xfd,0xd9,0x29,0x00,0x00,
  0x3b
};

/* Generated by reswrap from file copy.gif */
const unsigned char copydoc[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x0f,0x00,0xa1,0x00,0x00,0xc0,0xc0,0xc0,
  0x00,0x00,0x00,0xff,0xff,0xff,0x00,0x00,0x80,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,
  0x0f,0x00,0x00,0x02,0x38,0x44,0x8e,0xa0,0xcb,0x21,0x2f,0x0e,0x9b,0x0e,0xba,0x30,
  0xd7,0xbd,0xa3,0xfb,0x06,0x3d,0x43,0xd8,0x69,0x11,0x32,0x42,0xe3,0xa0,0x54,0xea,
  0x29,0x78,0xec,0x76,0xa4,0x61,0x6c,0xdc,0x23,0x72,0xe0,0x7c,0x7d,0x13,0x65,0x00,
  0x3b,0x1e,0x2e,0x63,0x23,0x0d,0x65,0xcc,0xa1,0xf3,0xe9,0x2c,0x00,0x00,0x3b
};

/* Generated by reswrap from file paste.gif */
const unsigned char pastedoc[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x0f,0x00,0xc2,0x00,0x00,0xc0,0xc0,0xc0,
  0x00,0x00,0x00,0xff,0xff,0x00,0x80,0x80,0x80,0x80,0x80,0x00,0x00,0x00,0x80,0xff,
  0xff,0xff,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,0x0f,0x00,0x00,0x03,
  0x47,0x08,0xba,0xd1,0xb1,0x10,0xb8,0x26,0xc4,0x7c,0x2a,0x0c,0x32,0x82,0xf5,0x01,
  0x47,0x60,0x61,0x17,0x49,0x1c,0xb9,0x5d,0xce,0x9a,0x89,0x5b,0x0c,0xab,0x70,0x61,
  0xdf,0x05,0x09,0x13,0x85,0xe1,0xfb,0xb6,0x8c,0x6c,0xd0,0xfb,0x19,0x7a,0x05,0x94,
  0x0c,0x69,0x3b,0x06,0x35,0x35,0xe3,0x2f,0x59,0x12,0x31,0x6f,0xc7,0xcc,0xa4,0x28,
  0x4d,0x9e,0x70,0xe0,0x93,0x78,0x91,0x00,0x00,0x3b
};

/* Generated by reswrap from file print.gif */
const unsigned char printdoc[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x0f,0x00,0xc2,0x00,0x00,0xc0,0xc0,0xc0,
  0x00,0x00,0x00,0xff,0xff,0xff,0x80,0x80,0x80,0xff,0xff,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,0x0f,0x00,0x00,0x03,
  0x35,0x08,0xaa,0xd1,0xee,0x6b,0x05,0x41,0x2b,0x0d,0x91,0x89,0x37,0x71,0x9e,0x56,
  0xe5,0x45,0xdd,0xb3,0x41,0x5a,0x78,0x01,0x58,0xcb,0x3d,0x2c,0xdb,0x64,0x5f,0xfc,
  0xde,0xde,0x9c,0x0d,0x3c,0x63,0xd3,0x84,0x20,0x63,0x86,0xe3,0xc4,0x8e,0xb4,0x9f,
  0xa4,0x38,0x4a,0x3a,0x01,0x09,0x00,0x3b
};





#endif
