// stc.cpp - Styled Text Control using the Scintilla cource code edit control
// implemented for the FOX toolkit
// SB: initial release; based on ScintillaWX by Robin Dunn.

#include <ctype.h>
#include <sys/time.h>
#include <unistd.h>
#include <fox/fx.h>
#include <fox/fxkeys.h>
#include "stc.h"
#include "resource.h"


// This is to prevent the compiler to optimize them away
int FXForceScintillaLexers(void) {
  extern LexerModule lmAda;
  extern LexerModule lmAVE;
  extern LexerModule lmConf;
  extern LexerModule lmCPP;
  extern LexerModule lmEiffel;
  extern LexerModule lmHTML;
  extern LexerModule lmLISP;
  extern LexerModule lmLua;
  extern LexerModule lmPascal;
  extern LexerModule lmPerl;
  extern LexerModule lmPython;
  extern LexerModule lmRuby;
  extern LexerModule lmSQL;
  extern LexerModule lmVB;
  //Others:
  extern LexerModule lmBatch;
  extern LexerModule lmDiff;
  extern LexerModule lmProps;
  extern LexerModule lmMake;
  extern LexerModule lmErrorList;
  extern LexerModule lmLatex;

  if (&lmAda && &lmAVE && &lmConf && &lmCPP && &lmEiffel && &lmHTML &&
      &lmLISP && &lmLua && &lmPascal && &lmPerl && &lmPython && &lmRuby &&
      &lmSQL && &lmVB && &lmBatch && &lmDiff && &lmProps && &lmMake &&
      &lmErrorList && &lmLatex) return 1;
  else return 0;
}



///////////////////////////////////////////////////////////////////////////////
//// Message map

FXDEFMAP(FXStyledText) FXStyledTextMap[]={

//[1] System messages:

  // Paint, Update, Resize:
  FXMAPFUNC(SEL_PAINT,                0,    FXStyledText::onPaint),
  FXMAPFUNC(SEL_UPDATE,               0,    FXStyledText::onUpdate),
  FXMAPFUNC(SEL_CONFIGURE,            0,    FXStyledText::onSize),

  // Key press and release:
  FXMAPFUNC(SEL_KEYPRESS,             0,    FXStyledText::onKeyPress),
  FXMAPFUNC(SEL_KEYRELEASE,           0,    FXStyledText::onKeyRelease),

  // Focus in and out (alt: FXStyledText::ID_EDIT)
  FXMAPFUNC(SEL_FOCUSIN,              0,    FXStyledText::onFocusIn),
  FXMAPFUNC(SEL_FOCUSOUT,             0,    FXStyledText::onFocusOut),

  // Left mouse button:
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,      0,    FXStyledText::onLeftBtnPress),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE,    0,    FXStyledText::onLeftBtnRelease),

  // Right mouse button:
  FXMAPFUNC(SEL_RIGHTBUTTONPRESS,     0,    FXStyledText::onRightBtnPress),
  FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,   0,    FXStyledText::onRightBtnRelease),

  // Midle mouse button:
  FXMAPFUNC(SEL_MIDDLEBUTTONPRESS,    0,    FXStyledText::onMiddleBtnPress),
  FXMAPFUNC(SEL_MIDDLEBUTTONRELEASE,  0,    FXStyledText::onMiddleBtnRelease),

  // Mouse motion:
  FXMAPFUNC(SEL_MOTION,               0,    FXStyledText::onMotion),

  // Selection:
  FXMAPFUNC(SEL_SELECTION_LOST,       0,    FXStyledText::onSelectionLost),
  FXMAPFUNC(SEL_SELECTION_GAINED,     0,    FXStyledText::onSelectionGained),
  FXMAPFUNC(SEL_SELECTION_REQUEST,    0,    FXStyledText::onSelectionRequest),

  // Clipboard:
  FXMAPFUNC(SEL_CLIPBOARD_LOST,       0,    FXStyledText::onClipboardLost),
  FXMAPFUNC(SEL_CLIPBOARD_GAINED,     0,    FXStyledText::onClipboardGained),
  FXMAPFUNC(SEL_CLIPBOARD_REQUEST,    0,    FXStyledText::onClipboardRequest),

	// DND:
  FXMAPFUNC(SEL_BEGINDRAG,            0,    FXStyledText::onBeginDrag),
  FXMAPFUNC(SEL_DRAGGED,              0,    FXStyledText::onDragged),
  FXMAPFUNC(SEL_ENDDRAG,              0,    FXStyledText::onEndDrag),
  FXMAPFUNC(SEL_DND_ENTER,            0,    FXStyledText::onDNDEnter),
  FXMAPFUNC(SEL_DND_LEAVE,            0,    FXStyledText::onDNDLeave),
  FXMAPFUNC(SEL_DND_MOTION,           0,    FXStyledText::onDNDMotion),
  FXMAPFUNC(SEL_DND_DROP,             0,    FXStyledText::onDNDDrop),
  FXMAPFUNC(SEL_DND_REQUEST,          0,    FXStyledText::onDNDRequest),



//[2] FOX messages:

  // Horizontal scrolling:
  FXMAPFUNC(SEL_COMMAND, FXWindow::ID_HSCROLLED, FXStyledText::onHScrollerChanged),
  FXMAPFUNC(SEL_CHANGED, FXWindow::ID_HSCROLLED, FXStyledText::onHScrollerDragged),

  // Vertical scrolling:
  FXMAPFUNC(SEL_COMMAND, FXWindow::ID_VSCROLLED, FXStyledText::onVScrollerChanged),
  FXMAPFUNC(SEL_CHANGED, FXWindow::ID_VSCROLLED, FXStyledText::onVScrollerDragged),

  // Blinker timer:
  FXMAPFUNC(SEL_TIMEOUT, FXWindow::ID_CARETBLINK, FXStyledText::onBlink),



//[3] Application specific messages:

  // Messages from Scintilla - Change, and Notify
  FXMAPFUNC(ScintillaFX::SCIMSG_CHANGE, ScintillaFX::ID_SCINTILLA, FXStyledText::onSciChange),
  FXMAPFUNC(ScintillaFX::SCIMSG_NOTIFY, ScintillaFX::ID_SCINTILLA, FXStyledText::onSciNotify),

  // Messages from the aplication - Editing commands

  // Insert character:
  FXMAPFUNC(SEL_COMMAND, FXStyledText::ID_INSERT_CHAR, FXStyledText::onCmdInsertChar),
  // Insert string:
  FXMAPFUNC(SEL_COMMAND, FXStyledText::ID_INSERT_STRING, FXStyledText::onCmdInsertString),
  // Goto:
  FXMAPFUNC(SEL_COMMAND, FXStyledText::ID_GOTO_LINE,   FXStyledText::onCmdGotoLine),
	// Undo:
	FXMAPFUNC(SEL_COMMAND, FXStyledText::ID_UNDO,        FXStyledText::onCmdUndo),
	// Redo:
	FXMAPFUNC(SEL_COMMAND, FXStyledText::ID_REDO,        FXStyledText::onCmdRedo),

	// View/hide LINENUM margin:
	FXMAPFUNC(SEL_COMMAND, FXStyledText::ID_LINENUM_MARGIN, FXStyledText::onCmdTogleMargin),
	FXMAPFUNC(SEL_UPDATE,  FXStyledText::ID_LINENUM_MARGIN, FXStyledText::onUpdMargin),
	// View/hide SYMBOLS margin:
	FXMAPFUNC(SEL_COMMAND, FXStyledText::ID_SYMBOL_MARGIN,  FXStyledText::onCmdTogleMargin),
	FXMAPFUNC(SEL_UPDATE,  FXStyledText::ID_SYMBOL_MARGIN,  FXStyledText::onUpdMargin),
	// View/hide FOLD margin:
	FXMAPFUNC(SEL_COMMAND, FXStyledText::ID_FOLD_MARGIN,    FXStyledText::onCmdTogleMargin),
  FXMAPFUNC(SEL_UPDATE,  FXStyledText::ID_FOLD_MARGIN,    FXStyledText::onUpdMargin),

  // Cut:
  FXMAPFUNC(SEL_COMMAND,  FXStyledText::ID_CUT_SEL,    FXStyledText::onCmdCutSel),
  //FXMAPFUNC(SEL_UPDATE, FXStyledText::ID_CUT_SEL,    FXStyledText::onUpdHaveSelection),
  // Copy:
  FXMAPFUNC(SEL_COMMAND,  FXStyledText::ID_COPY_SEL,   FXStyledText::onCmdCopySel),
  //FXMAPFUNC(SEL_UPDATE, FXStyledText::ID_COPY_SEL,   FXStyledText::onUpdHaveSelection),
  // Paste:
  FXMAPFUNC(SEL_COMMAND,  FXStyledText::ID_PASTE_SEL,  FXStyledText::onCmdPasteSel),
  //FXMAPFUNC(SEL_UPDATE, FXStyledText::ID_PASTE_SEL,  FXStyledText::onUpdYes),
  // Delete:
  FXMAPFUNC(SEL_COMMAND,  FXStyledText::ID_DELETE_SEL, FXStyledText::onCmdDeleteSel),
  //FXMAPFUNC(SEL_UPDATE, FXStyledText::ID_DELETE_SEL, FXStyledText::onUpdHaveSelection),
  // Select All:
  FXMAPFUNC(SEL_COMMAND,  FXStyledText::ID_SELECTALL,  FXStyledText::onCmdSelectAll),



};


// Object implementation
FXIMPLEMENT(FXStyledText, FXScrollArea, FXStyledTextMap, ARRAYNUMBER(FXStyledTextMap))


// For deserialization
FXStyledText::FXStyledText(){
  // todo: fill up
  blinker = NULL;
  flags|=FLAG_ENABLED;  // Do we need maintain the 'flags' as the FXText does?
                        // At min. the FLAG_ENABLED is required !!
  clipbuffer=NULL;
  cliplength=0;
}


// Text widget
FXStyledText::FXStyledText(FXComposite *p, FXObject* tgt, FXSelector sel, FXuint opts,
                           FXint x, FXint y, FXint w, FXint h) :
              FXScrollArea(p, opts, x, y, w, h),
              delimiters("~.,/\\`'!@#$%^&*()-=+{}|[]\":;<>?")
{

  font          = getApp()->getNormalFont();
  textColor     = getApp()->getForeColor();
  defaultCursor = getApp()->getDefaultCursor(DEF_TEXT_CURSOR);     //private
	dragCursor    = getApp()->getDefaultCursor(DEF_TEXT_CURSOR);     //private
  target        = tgt;
  message       = sel;
  blinker       = NULL;
  m_BlinkSpeed  = getApp()->getBlinkSpeed() /4;
  clipbuffer    = NULL;
  cliplength    = 0;
  mode          = MOUSE_NONE;

  flags|=FLAG_ENABLED;  // Do we need maintain the 'flags' as the FXText does?
                        // At min. the FLAG_ENABLED is required !!

  m_MarginSize[0] = 0;
	m_MarginSize[1] = 0;
	m_MarginSize[2] = 0;

  m_arrow1 = new FXCursor(getApp(),arrow1_b,arrow1_m,arrow1_w,arrow1_h,arrow1_x,arrow1_y);
  m_arrow2 = new FXCursor(getApp(),arrow2_b,arrow2_m,arrow2_w,arrow2_h,arrow2_x,arrow2_y);
  m_arrow4 = new FXCursor(getApp(),arrow4_b,arrow4_m,arrow4_w,arrow4_h,arrow4_x,arrow4_y);

  //setScrollStyle(SCROLLERS_NORMAL|HSCROLLING_ON|VSCROLLING_ON);
  setScrollStyle(HSCROLLER_ALWAYS|HSCROLLING_ON|VSCROLLER_ALWAYS|VSCROLLING_ON);
}


// Create window
void FXStyledText::create(){
  FXScrollArea::create();
	font->create();
  m_arrow1->create();
  m_arrow2->create();
  m_arrow4->create();
  m_sfx = new ScintillaFX(this);

  LoadConfig();

  if(!deleteType){deleteType=getApp()->registerDragType(deleteTypeName);}
  if(!textType){textType=getApp()->registerDragType(textTypeName);}
  dropEnable();
  recalc();

  m_nitems = 1;  // there is always one line
  m_itemHeight = font->getFontHeight();

  horizontal->setLine(font->getTextWidth(" ",1));
  vertical->setLine(m_itemHeight);
}


// dtor
FXStyledText::~FXStyledText(){
  delete m_sfx;
  font=(FXFont*)-1;
  if(blinker) getApp()->removeTimeout(blinker);
  blinker=(FXTimer*)-1;
  delete m_arrow1;
  delete m_arrow2;
  delete m_arrow4;
  //more?
}


///////////////////////////////////////////////////////////////////////////////
//// For communicating with Scintilla proper

long FXStyledText::SendMsg(int msg, long wp, long lp) {
  return m_sfx->WndProc(msg, wp, lp);
}


///////////////////////////////////////////////////////////////////////////////
//// Messages:


// Scintilla notification:

long FXStyledText::onSciChange(FXObject*, FXSelector, void* ptr){
  //printf("Got Sci Change notification\n");
}

long FXStyledText::onSciNotify(FXObject*, FXSelector, void* ptr){
  SCNotification* scn = (SCNotification* )ptr;

// scn:
//
// struct NotifyHeader nmhdr;    // uint code, uint idFrom, void* hwndFrom
// int position;			           // SCN_STYLENEEDED, SCN_MODIFIED
// int ch;					             // SCN_CHARADDED, SCN_KEY
// int modifiers;			           // SCN_KEY
// int modificationType;	       // SCN_MODIFIED
// const char *text;		         // SCN_MODIFIED
// int length;				           // SCN_MODIFIED
// int linesAdded;	             // SCN_MODIFIED
// int line;			               // SC_MOD_CHANGEFOLD
// int foldLevelNow;	           // SC_MOD_CHANGEFOLD
// int foldLevelPrev;	           // SC_MOD_CHANGEFOLD
// int margin;	                 // SCN_MARGINCLICK
// #ifdef MACRO_SUPPORT
// int message;	                 // SCN_MACRORECORD
// int wParam;	                 // SCN_MACRORECORD
// int lParam;	                 // SCN_MACRORECORD

// scn->modificationType:
//
// #define SC_MOD_INSERTTEXT     0x1
// #define SC_MOD_DELETETEXT     0x2
// #define SC_MOD_CHANGESTYLE    0x4
// #define SC_MOD_CHANGEFOLD     0x8
// #define SC_PERFORMED_USER     0x10
// #define SC_PERFORMED_UNDO     0x20
// #define SC_PERFORMED_REDO     0x40
// #define SC_LASTSTEPINUNDOREDO 0x100
// #define SC_MOD_CHANGEMARKER   0x200
// #define SC_MOD_BEFOREINSERT   0x400
// #define SC_MOD_BEFOREDELETE   0x800

// scn->nmhdr.code:
//
// #define SCN_STYLENEEDED       2000
// #define SCN_CHARADDED         2001
// #define SCN_SAVEPOINTREACHED  2002
// #define SCN_SAVEPOINTLEFT     2003
// #define SCN_MODIFYATTEMPTRO   2004
// #define SCN_KEY               2005
// #define SCN_DOUBLECLICK       2006
// #define SCN_UPDATEUI          2007
// #define SCN_CHECKBRACE        2007
// #define SCN_MODIFIED          2008
// #define SCN_MACRORECORD       2009
// #define SCN_MARGINCLICK       2010
// #define SCN_NEEDSHOWN         2011
// #define SCN_POSCHANGED        2012
// #define SCN_PAINTED           2013


  //printf("-->Sci Notification: nmhdr.code= %d, modificationType= 0x%X\n",
  //       scn->nmhdr.code, scn->modificationType);

  //switch (scn->nmhdr.code) {...
}


// Fox scrolling messages
// SB: I think we should do the "if target->handle(...)" in these...

long FXStyledText::onHScrollerDragged(FXObject*, FXSelector, void* ptr) {
  FXint new_x = (FXint )(long )ptr;
  if (new_x == -pos_x) {
    //printf("H_DRAGGED: skipping position %d!\n", new_x);
    return 1;
  }
  m_sfx->DoHScroll(new_x);
  //printf("H_DRAGGED: current pos_x= %d\n", -pos_x);
  return 1;
}

//needed?
long FXStyledText::onHScrollerChanged(FXObject*, FXSelector, void* ptr) {
  FXint new_x = (FXint )(long )ptr;
  if (new_x == -pos_x) {
    //printf("H_DRAGGED: skipping position %d!\n", new_x);
    return 1;
  }
  m_sfx->DoHScroll(new_x);
  //printf("H_DRAGGED: current pos_x= %d\n", -pos_x);
  return 1;
}


long FXStyledText::onVScrollerDragged(FXObject* obj, FXSelector sel, void* ptr) {
  FXint new_y = (FXint )(long )ptr;
  if (new_y == -pos_y) {
    //printf("V_DRAGGED: skipping position %d!\n", new_y);
    return 1;
  }
  FXint line_h = vertical->getLine();
  FXint line_y = new_y / line_h;
  m_sfx->DoVScroll(line_y);                      // on dragging and scrolling
  //printf("V_DRAGGED: current line= %d, new_y= %d pos_y= %d\n", line_y, new_y, -pos_y);
  return 1;
}


long FXStyledText::onVScrollerChanged(FXObject* obj, FXSelector sel, void* ptr) {
  FXint new_y = (FXint )(long )ptr;
  if (new_y == -pos_y) {
    //printf("V_CHANGED: skipping position %d!\n", new_y);
    return 1;
  }
  FXint line_h = vertical->getLine();
  FXint line_y = new_y / line_h;
  m_sfx->DoVScroll(line_y);                      // on release when dragging
  //printf("V_CHANGED: current line= %d, new_y= %d pos_y= %d\n", line_y, new_y, -pos_y);
  return 1;
}


///////////////////////////////////////////////////////////////////////////////
//// DND
// See FXTreeList.cpp, FXDirList, and FXText.cpp

long FXStyledText::onBeginDrag(FXObject* obj, FXSelector sel, void* ptr) {
//printf("stc: onBeginDrag() called\n");
//flags|=FLAG_SCROLLINSIDE;

  grab();
  if(FXScrollArea::onBeginDrag(obj,sel,ptr)) return 1;
  beginDrag(&textType, 1);
  return 1;
}


long FXStyledText::onDragged(FXObject* obj, FXSelector sel, void* ptr) {
//printf("stc: onDragged() called\n");

  if(FXScrollArea::onDragged(obj,sel,ptr)) return 1;

  FXEvent* ev=(FXEvent*)ptr;
  FXDragAction action=DRAG_COPY;

  //if(isEditable()){
    if(isDropTarget()) action=DRAG_MOVE;
    if(ev->state&CONTROLMASK) action=DRAG_COPY;
    if(ev->state&SHIFTMASK) action=DRAG_MOVE;
  //}

  handleDrag(ev->root_x, ev->root_y, action);
  SetDNDCursor(ptr);
  return 1;
}


void FXStyledText::SetDNDCursor(void* ptr) {
  FXEvent* ev=(FXEvent*)ptr;
  FXDragAction action=DRAG_COPY;

  //if(isEditable()){
    if(isDropTarget()) action=DRAG_MOVE;
    if(ev->state&CONTROLMASK) action=DRAG_COPY;
    if(ev->state&SHIFTMASK) action=DRAG_MOVE;
  //}

  if(didAccept()!=DRAG_REJECT){
    if(action==DRAG_COPY) setDragCursor(m_arrow1);  // copy
    else setDragCursor(m_arrow2);                   // move
  } else setDragCursor(m_arrow4);                   // reject

}


long FXStyledText::onEndDrag(FXObject* obj, FXSelector sel, void* ptr) {
//printf("stc: onEndDrag() called\n");

  if(FXScrollArea::onEndDrag(obj,sel,ptr)) return 1;

  FXbool accepted = (didAccept() != DRAG_REJECT);
  if (!accepted) getApp()->beep();

  endDrag(accepted);
  setDragCursor(getApp()->getDefaultCursor(DEF_TEXT_CURSOR));
  return 1;
}


long FXStyledText::onDNDEnter(FXObject* obj, FXSelector sel, void* ptr) {
//printf("stc: onDNDEnter() called\n");
  FXScrollArea::onDNDEnter(obj, sel, ptr);
  return 1;
}


long FXStyledText::onDNDLeave(FXObject* obj, FXSelector sel, void* ptr) {
//printf("stc: onDNDLeave() called\n");
  //flags&=~FLAG_SCROLLINSIDE;
  FXScrollArea::onDNDLeave(obj, sel, ptr);
  return 1;
}


long FXStyledText::onDNDMotion(FXObject* obj, FXSelector sel, void* ptr) {
//printf("DNDMotion() called\n");
  if(FXScrollArea::onDNDMotion(obj, sel, ptr)) return 1;

  FXEvent* ev=(FXEvent*)ptr;
  Point pt;
  pt.x = ev->win_x; pt.y = ev->win_y;

  // Correct drop type
  if(offeredDNDType(FROM_DRAGNDROP, textType) /*&& isEditable*/) {

    FXDragAction action = inquireDNDAction();
    if(action==DRAG_COPY || action==DRAG_MOVE) {
      if(!IsPointInSelection(FXPointFromPoint(pt)) &&
         !IsPointInSelMargin(FXPointFromPoint(pt)) &&
         (pt.y < viewport_h) ) {
        acceptDrop(DRAG_ACCEPT);
        SetDNDCursor(ptr);
        return 1;
      }

    }
  }

  return 0;
}


long FXStyledText::onDNDDrop(FXObject* obj, FXSelector sel, void* ptr) {
// Called when something drops inside
//printf("stc: DNDDrop() called\n");

  ungrab();  // <--- ???
  if (FXScrollArea::onDNDDrop(obj,sel,ptr)) return 1;
  FXEvent* ev=(FXEvent*)ptr;

  Point pt;
  pt.x = ev->win_x; pt.y = ev->win_y;
  FXuchar *data, *junk;
  FXuint len=0, dum=0;

  if(inquireDNDAction()==DRAG_MOVE){
    getDNDData(FROM_DRAGNDROP, deleteType, junk, dum);
    FXASSERT(!junk);
  }

  if(getDNDData(FROM_DRAGNDROP, textType, data, len)){
    FXRESIZE(&data, FXchar, len+1); data[len]='\0';
    FXString text; text.format("%s",data);
    m_sfx->DoDNDDrop(pt, text);
    FXFREE(&data);
   }

  setDragCursor(getApp()->getDefaultCursor(DEF_TEXT_CURSOR));
  return 1;
}


long FXStyledText::onDNDRequest(FXObject* obj, FXSelector sel, void* ptr) {
//printf("DNDRequest() called\n");

  if(FXScrollArea::onDNDRequest(obj,sel,ptr)) return 1;
  FXEvent* ev=(FXEvent*)ptr;

  // Delete the source
  if(ev->target==deleteType){
    ClearSelection();
    return 1;
  }

  if(ev->target==textType){
    if(!m_dragtext.empty()){
      FXuchar *data;
      FXuint len = m_dragtext.length();
      FXMEMDUP(&data, FXuchar, m_dragtext.text(), len);
      setDNDData(FROM_DRAGNDROP, ev->target, data, len);
			m_dragtext = FXString::null;
    }
    return 1;
  }

  return 0;
}


///////////////////////////////////////////////////////////////////////////////
// FOX Events:

long FXStyledText::onPaint(FXObject*, FXSelector, void* ptr){
//printf("stc: Painting!\n");

  FXEvent* event=(FXEvent*)ptr;
  FXDCWindow dc(this,event);

  //printf("onPaint() x= %d, y= %d, w= %d, h= %d\n",
  //       event->rect.x, event->rect.y, event->rect.w, event->rect.h);

  dc.setTextFont(font);
  m_itemHeight = font->getFontHeight();  // update in case the font changed
                                         // TODO get the value from Editor::vs.lineHeight

  m_sfx->DoPaint(this, event->rect);
  return 1;
}


long FXStyledText::onUpdate(FXObject* sender,FXSelector sel,void* ptr){
  FXScrollArea::onUpdate(sender,sel,ptr);  //needed?
  return 1;
}


long FXStyledText::onSize(FXObject* sender,FXSelector sel,void* ptr){
  //FXEvent* event=(FXEvent*)ptr;
  //FXRectangle rec;
  //rec.x = event->rect.x; rec.y = event->rect.y;
  //rec.w = event->rect.w; rec.h = event->rect.h;

  m_sfx->DoSize();
  flags|=FLAG_DIRTY;
  return 1;
}


///////////////////////////////////////////////////////////////////////////////
//// Keys

long FXStyledText::onKeyPress(FXObject* sender,FXSelector sel,void* ptr){
  //StopBlinker();                  // Editor will start it

  FXEvent* event=(FXEvent*)ptr;
  char entry;
  entry=fxkeyval(event->code, event->state);
  //printf("stc: onKeyPress() key_char= %c\n", entry);

  FXbool shift=0, ctrl=0, alt=0;

  if (event->state & SHIFTMASK)    shift=1;
  if (event->state & CONTROLMASK)  ctrl=1;
  if (event->state & ALTMASK)      alt=1;

  m_sfx->DoKeyDown(event->code, shift, ctrl, alt);

  if (event->code < 255) {          // sufficient?
    handle(this, MKUINT(ID_INSERT_CHAR, SEL_COMMAND), (void*)(long)entry);
  }

  return 1;
}


long FXStyledText::onKeyRelease(FXObject* obj,FXSelector sel,void* ptr){
  return 0;
}


///////////////////////////////////////////////////////////////////////////////
// Insert CHAR, STRING

long FXStyledText::onCmdInsertChar(FXObject* sender,FXSelector sel,void* ptr){
  FXchar ch=(FXchar)(long)ptr;
  //printf("stc: inserting: %c\n", ch);
  //StopBlinker();                  // Editor will start it

  BeginUndoAction();
  m_sfx->DoAddChar(ch);
  EndUndoAction();
  return 1;
}


long FXStyledText::onCmdInsertString(FXObject* sender,FXSelector sel,void* ptr){
//printf("stc: inserting: %s\n", s);
  FXchar* s = (FXchar*)(long)ptr;
  //if(!isEditable()) return 1;
  FXint pos = GetCurrentPos();
  InsertText(pos, s);
  return 1;
}


///////////////////////////////////////////////////////////////////////////////
//// Focus, blinker

long FXStyledText::onFocusIn(FXObject* sender, FXSelector sel, void* ptr){
//printf("stc: onFocusIn() entered\n");
  flags |= FLAG_FOCUSED;
  FXScrollArea::onFocusIn(sender, sel, ptr);
  m_sfx->DoGainFocus();
  return 1;
}


long FXStyledText::onFocusOut(FXObject* sender, FXSelector sel, void* ptr){
//printf("stc: onFocusOut() entered\n");
  flags &= ~FLAG_FOCUSED;
  FXScrollArea::onFocusOut(sender, sel, ptr);
  m_sfx->DoLoseFocus();
  return 1;
}


long FXStyledText::onBlink(FXObject* sender, FXSelector sel, void* ptr){
//printf("stc: onBlink() entered()\n");
  blinker = getApp()->addTimeout(m_BlinkSpeed, this, FXWindow::ID_CARETBLINK);
  m_sfx->DoTick();
  return 0;
}


///////////////////////////////////////////////////////////////////////////////
//// Mouse

void FXStyledText::MouseDown(FXObject* sender, FXSelector sel, void* ptr) {

  ////grab();
  FXEvent* ev=(FXEvent*)ptr;
  Point pt;
  pt.x = ev->win_x; pt.y = ev->win_y;

  FXbool shift=0, ctrl=0, alt=0;

  if (ev->state & SHIFTMASK)    shift=1;
  if (ev->state & CONTROLMASK)  ctrl=1;
  if (ev->state & ALTMASK)      alt=1;

//printf("stc: MouseDown() pt= (%d, %d), time= %u, shift/ctrl/alt= %d/%d/%d\n",
//       pt.x, pt.y, ev->time, shift, ctrl, alt);

  m_sfx->DoButtonDown(pt, ev->time, shift, ctrl, alt);
}


void FXStyledText::MouseUp(FXObject* sender,FXSelector sel,void* ptr) {
  FXEvent* ev=(FXEvent*)ptr;
  Point pt;
  pt.x = ev->win_x; pt.y = ev->win_y;
  FXbool shift=0, ctrl=0, alt=0;

  if (ev->state & SHIFTMASK)    shift=1;
  if (ev->state & CONTROLMASK)  ctrl=1;
  if (ev->state & ALTMASK)      alt=1;

//printf("stc: MouseUp() pt= (%d, %d), time= %u, shift/ctrl/alt= %d/%d/%d\n",
//       pt.x, pt.y, ev->time, shift, ctrl, alt);

  m_sfx->DoButtonUp(pt, ev->time, ctrl, ev->moved);
}


// Pressed left button
long FXStyledText::onLeftBtnPress(FXObject* obj, FXSelector sel, void* ptr) {
//printf("stc: onLeftBtnPress() entered\n");

  FXEvent* ev=(FXEvent*)ptr;
  FXint pos;
	flags|=FLAG_PRESSED;
  flags&=~FLAG_TIP;

  handle(this, MKUINT(0, SEL_FOCUS_SELF), ptr);

  if(target && target->handle(this, MKUINT(message, SEL_LEFTBUTTONPRESS), ptr))
    return 1;

  //if(isEnabled()){
     MouseDown(obj, sel, ptr);
     return 1;
  //}

  return 0;
}


// Released left button
long FXStyledText::onLeftBtnRelease(FXObject* obj,FXSelector sel,void* ptr) {
//printf("stc: onLeftBtnRelease() entered\n");

/*
  if(isEnabled()){
    ungrab();
    flags&=~FLAG_PRESSED;
    flags|=FLAG_UPDATE;
    stopAutoScroll();
    if(target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONRELEASE),ptr)) return 1;
    dragmode=MOUSE_NONE;
    return 1;
  }
  return 0;
*/

  //StartBlinker();

  FXuint flg=flags;
  flags|=FLAG_UPDATE;
  flags&=~(FLAG_PRESSED|FLAG_TRYDRAG|FLAG_DODRAG);

  // Was dragging
  if(flg&FLAG_DODRAG){
    handle(this,MKUINT(0,SEL_ENDDRAG),ptr);
  }

  MouseUp(obj, sel, ptr);
  return 1;
}


// Pressed right button
long FXStyledText::onRightBtnPress(FXObject*,FXSelector,void* ptr){
//printf("stc: onRightBtnPress() entered\n");

/*
  FXEvent* ev=(FXEvent*)ptr;
  flags&=~FLAG_TIP;
  handle(this,MKUINT(0,SEL_FOCUS_SELF),ptr);
  if(isEnabled()){
    grab();
    if(target && target->handle(this,MKUINT(message,SEL_RIGHTBUTTONPRESS),ptr)) return 1;
    flags&=~FLAG_UPDATE;
    flags|=FLAG_PRESSED|FLAG_SCROLLING;
    grabx=ev->win_x-pos_x;
    graby=ev->win_y-pos_y;
    return 1;
  }
  return 0;
*/

  handle(this, MKUINT(0, SEL_FOCUS_SELF), ptr);
  return 0;
}


// Released right button
long FXStyledText::onRightBtnRelease(FXObject*,FXSelector,void* ptr){
//printf("stc: onRightBtnRelease() entered\n");

/*
  if(isEnabled()){
    ungrab();
    flags&=~(FLAG_PRESSED|FLAG_SCROLLING);
    flags|=FLAG_UPDATE;
    if(target && target->handle(this,MKUINT(message,SEL_RIGHTBUTTONRELEASE),ptr)) return 1;
    return 1;
  }
  return 0;
*/

  FXEvent* ev=(FXEvent*)ptr;
  Point pt;
  pt.x = ev->root_x; pt.y = ev->root_y;

  if(!ev->moved){
    m_pane = new FXMenuPane(this);
    m_sfx->DoContextMenu(pt);
    m_pane->create();
    m_pane->popup(NULL, pt.x, pt.y);
    getApp()->runModalWhileShown(m_pane);
  }

  return 1;
}


long FXStyledText::onMiddleBtnPress(FXObject* obj, FXSelector sel, void* ptr) {

  FXEvent* ev = (FXEvent*)ptr;
  FXuchar *data;
  FXuint len;
  flags&=~FLAG_TIP;

  handle(this, MKUINT(0, SEL_FOCUS_SELF), ptr);

  //if(isEditable(){

      if(target && target->handle(this,MKUINT(message,SEL_MIDDLEBUTTONPRESS),ptr))
        return 1;

      if(getDNDData(FROM_SELECTION, stringType, data, len)){
        FXRESIZE(&data, FXchar, len +1); data[len]='\0';
        //FXString text; text.format("%s",data);

        FXPoint pt;
        pt.x = ev->win_x; pt.y = ev->win_y;
        FXint pos = PositionFromPoint(pt);
        GotoPos(pos);

        handle(this, MKUINT(ID_INSERT_STRING, SEL_COMMAND), (void*)data);
        //InsertText(pos, text);

        FXFREE(&data);
        flags&=~FLAG_UPDATE;
        return 1;
      }

  //}

  return 0;
}


// Released middle button
long FXStyledText::onMiddleBtnRelease(FXObject* obj, FXSelector sel, void* ptr){
  return 0;
}


// Handle real or simulated mouse motion
long FXStyledText::onMotion(FXObject* obj,FXSelector sel,void* ptr){

  FXEvent* ev=(FXEvent*)ptr;
  Point pt;
  pt.x = ev->win_x; pt.y = ev->win_y;
  //printf("stc: onMotion() entered pt.y= %d\n", pt.y);

  // Drag and drop mode
  if(flags&FLAG_DODRAG){
    handle(this, MKUINT(0, SEL_DRAGGED), ptr);
    //return 1;
  }

  // Tentative drag and drop
  if((flags&FLAG_TRYDRAG) && ev->moved){
    flags&=~FLAG_TRYDRAG;
    if(handle(this, MKUINT(0, SEL_BEGINDRAG), ptr)){
      flags|=FLAG_DODRAG;
    }
    //return 1;
  }

  m_sfx->DoButtonMove(pt);
  return 1;
}


///////////////////////////////////////////////////////////////////////////////
//// Utilities

// Load default configuration (temp)

void FXStyledText::LoadConfig() {
  SetLexer(SCLEX_CPP);

  FXColor col_comment       = FXRGB(128,128,0);
  FXColor col_number        = FXRGB(128,0,128);
  FXColor col_word          = FXRGB(0,0,255);
  FXColor col_default       = FXRGB(0,0,0);
  FXColor col_string        = FXRGB(255,0,255);
  FXColor col_preprocessor  = FXRGB(255,0,0);
  FXColor col_identifier    = FXRGB(0,0,0);
  FXColor col_operator      = FXRGB(0,0,0);

  FXString keywords = "asm auto bool break case catch char class const const_cast continue \
default delete do double dynamic_cast else enum explicit export extern false float for \
friend goto if inline int long mutable namespace new operator private protected public \
register reinterpret_cast return short signed sizeof static static_cast struct switch \
template this throw true try typedef typeid typename union unsigned using \
virtual void volatile wchar_t while";

  SendMsg(SCI_SETKEYWORDS, 0, (long)keywords.text());

  // see SciLexer.h
  SendMsg(SCI_STYLESETFORE, SCE_C_DEFAULT, col_default);           // SCE_C code: 0
  SendMsg(SCI_STYLESETFORE, SCE_C_COMMENT, col_comment);           // SCE_C code: 1
  SendMsg(SCI_STYLESETFORE, SCE_C_COMMENTLINE, col_comment);       // SCE_C code: 2
  SendMsg(SCI_STYLESETFORE, SCE_C_COMMENTDOC, col_comment);        // SCE_C code: 3
  SendMsg(SCI_STYLESETFORE, SCE_C_NUMBER, col_number);             // SCE_C code: 4
  SendMsg(SCI_STYLESETFORE, SCE_C_WORD, col_word);                 // SCE_C code: 5
  SendMsg(SCI_STYLESETFORE, SCE_C_STRING, col_string);             // SCE_C code: 6
  SendMsg(SCI_STYLESETFORE, SCE_C_CHARACTER, col_string);          // SCE_C code: 7
  SendMsg(SCI_STYLESETFORE, SCE_C_UUID, col_default);              // SCE_C code: 8
  SendMsg(SCI_STYLESETFORE, SCE_C_PREPROCESSOR, col_preprocessor); // SCE_C code: 9
  SendMsg(SCI_STYLESETFORE, SCE_C_OPERATOR, col_operator);         // SCE_C code: 10
  SendMsg(SCI_STYLESETFORE, SCE_C_IDENTIFIER, col_identifier);     // SCE_C code: 11
  SendMsg(SCI_STYLESETFORE, SCE_C_STRINGEOL, col_default);         // SCE_C code: 12
  SendMsg(SCI_STYLESETFORE, SCE_C_VERBATIM, col_default);          // SCE_C code: 13
  SendMsg(SCI_STYLESETFORE, SCE_C_REGEX, col_default);             // SCE_C code: 14

  SetFont("fixed", 12);

	// The 'line numbers' margin
  SetMarginType(0, SC_MARGIN_NUMBER);
  SetMarginSize(0, 3);
  EnableMargin(0, false);

  // The 'symbols' margin
  SetMarginType(1, SC_MARGIN_SYMBOL);
  SetMarginSize(1, 2);
  EnableMargin(1, false);

	// The 'fold' margins
  SetMarginType(2, SC_MARGIN_SYMBOL);
	SetMarginMask(2, SC_MASK_FOLDERS);
  SetMarginSize(2, 2);
  EnableMargin(2, false);

	//SendMsg(SCI_SETCARETWIDTH, 2);
  //getApp()->setScrollSpeed(10);
  m_sfx->SetCaretMargin(1);
	SetUndoCollection(true);
}


// Set font
void FXStyledText::SetFont(const FXString& face, int size) {
	for (int style = 0; style <= STYLE_MAX; style++) {
			StyleSetFace(style, face);
			StyleSetSize(style, size);
		}
	  font = new FXFont(getApp(), face, size);  //ok?
	  font->create();
    m_itemHeight = font->getFontHeight();  //TODO: use Editor::vs.lineHeight !
    layout();
}


// Start/stop blinker

long FXStyledText::StartBlinker() {
//printf("stc: StartBlinker() entered\n");
  if (blinker) return 0;
  blinker = getApp()->addTimeout(m_BlinkSpeed, this, FXWindow::ID_CARETBLINK);
  return 1;
}

long FXStyledText::StopBlinker() {
//printf("stc: StopBlinker() entered\n");
  //blinker is set to NULL
  blinker = getApp()->removeTimeout(blinker);
  return 1;
}


///////////////////////////////////////////////////////////////////////////////
// FXSrollArea and FXWindows virtuals
// Warning: if implementing getViewport... functions test them using the
// Window::GetClientPosition() in PlatFOX.cpp - last time it didn't work...

// Viewport height and width - from FXScrollArea (ok?)

FXint FXStyledText::getViewportHeight() {
//viewport_h = FXScrollArea::getViewportHeight() - font->getFontHeight();
  register FXint sh_h=0;
  register FXint sv_w=0;
  int _w, _h;

  _w=FXScrollArea::getViewportWidth();
  _h=FXScrollArea::getViewportHeight();

  // Get dimensions of the scroll bars
  if(!(options&HSCROLLER_NEVER)) sh_h=horizontal->getDefaultHeight();
  if(!(options&VSCROLLER_NEVER)) sv_w=vertical->getDefaultWidth();

  // Should we disable the scroll bars?
  // A bit tricky as the scrollbars may influence each other's presence
  if(!(options&(HSCROLLER_ALWAYS|VSCROLLER_ALWAYS)) &&
      (content_w<=_w) && (content_h<=_h)){sh_h=sv_w=0;}
  if(!(options&HSCROLLER_ALWAYS) && (content_w<=_w-sv_w)) sh_h=0;
  if(!(options&VSCROLLER_ALWAYS) && (content_h<=_h-sh_h)) sv_w=0;
  if(!(options&HSCROLLER_ALWAYS) && (content_w<=_w-sv_w)) sh_h=0;

  // Viewport size with scroll bars taken into account
  _w-=sv_w;
  _h-=sh_h;

  //printf ("stc: viewport_h=%d\n", _h);
  viewport_h = _h;
  return _h;
}


FXint FXStyledText::getViewportWidth() {
  register FXint sh_h=0;
  register FXint sv_w=0;
  int _w, _h;

  _w=FXScrollArea::getViewportWidth();
  _h=FXScrollArea::getViewportHeight();

  // Get dimensions of the scroll bars
  if(!(options&HSCROLLER_NEVER)) sh_h=horizontal->getDefaultHeight();
  if(!(options&VSCROLLER_NEVER)) sv_w=vertical->getDefaultWidth();

  // Should we disable the scroll bars?
  // A bit tricky as the scrollbars may influence each other's presence
  if(!(options&(HSCROLLER_ALWAYS|VSCROLLER_ALWAYS)) &&
    (content_w<=_w) && (content_h<=_h)){sh_h=sv_w=0;}
  if(!(options&HSCROLLER_ALWAYS) && (content_w<=_w-sv_w)) sh_h=0;
  if(!(options&VSCROLLER_ALWAYS) && (content_h<=_h-sh_h)) sv_w=0;
  if(!(options&HSCROLLER_ALWAYS) && (content_w<=_w-sv_w)) sh_h=0;

  // Viewport size with scroll bars taken into account
  _w-=sv_w;
  _h-=sh_h;

  //printf ("stc: viewport_w= %d\n", _w);
  viewport_w = _w;
  return _w;
}


//// Content width and height

FXint FXStyledText::getContentWidth() {
//printf("getContentWidth(): stc_width= %d, viewport_w= %d", getWidth(), viewport_w);
  return 2000;
}


FXint FXStyledText::getContentHeight() {
//printf("getContentHeight(): stc_height= %d, viewport_h= %d, content_w= %d\n",
//       getHeight(), viewport_h, m_nitems * m_itemHeight);
//content_h = (m_nitems +1) * m_itemHeight -1;  //OLD!!
  return m_nitems * m_itemHeight;
}


//// FXWindow virtuals

// Can focus?
FXbool FXStyledText::canFocus() const { return 1; }


// Enable the window
void FXStyledText::enable(){
  if(!(flags&FLAG_ENABLED)){
    FXScrollArea::enable();
    update(0,0,viewport_w,viewport_h);
  }
}


// Disable the window
void FXStyledText::disable(){
  if(flags&FLAG_ENABLED){
    FXScrollArea::disable();
    update(0,0,viewport_w,viewport_h);
  }
}


// Propagate size change
void FXStyledText::recalc(){
  FXScrollArea::recalc();
  flags|=FLAG_RECALC;
}


///////////////////////////////////////////////////////////////////////////////
// SB: on KDE we have focus problem - after a dialog or a menu popup is closed
// the focus is not restored to the canvas. Workaround: SEL_FOCUS_SELF.

long FXStyledText::onCmdGotoLine(FXObject* obj, FXSelector sel, void* ptr){
  FXint row=1;
  if(FXInputDialog::getInteger(row,(FXWindow*)target,"Goto Line","&Goto line number:",NULL,1,2147483647)){
    GotoLine(row -1);
  }
	handle(this, MKUINT(0, SEL_FOCUS_SELF), ptr);
  return 1;
}


long FXStyledText::onCmdUndo(FXObject*, FXSelector, void* ptr) {
  Undo();
  handle(this, MKUINT(0, SEL_FOCUS_SELF), ptr);
	return 1;
}


long FXStyledText::onCmdRedo(FXObject*, FXSelector, void* ptr) {
  Redo();
  handle(this, MKUINT(0, SEL_FOCUS_SELF), ptr);
	return 1;
}


///////////////////////////////////////////////////////////////////////////////
/////// Margin

void FXStyledText::EnableMargin(FXint margin, FXbool enable) {
  if (margin <0 || margin >3) return;
	if (enable) SetMarginWidth(margin, m_MarginSize[margin]);
	else SetMarginWidth(margin, 0);
}


FXbool FXStyledText::IsEnabledMargin(FXint margin) {
  return (GetMarginWidth(margin) > 0) ? true : false;
}


FXint FXStyledText::SetMarginSize(FXint margin, FXint chars) {
  if (margin <0 || margin >3) return 0;
  m_MarginSize[margin] = chars * font->getTextWidth(" ",1) +4;
  return GetMarginWidth(margin);  //the old value!
}


long FXStyledText::onCmdTogleMargin(FXObject* obj, FXSelector sel, void* ptr) {
  FXint id = SELID(sel);
	FXint margin;
	switch (id) {
	  case ID_LINENUM_MARGIN: margin = 0; break;
	  case ID_SYMBOL_MARGIN:  margin = 1; break;
    case ID_FOLD_MARGIN:    margin = 2; break;
	}
	FXbool enabled = IsEnabledMargin(margin);
  EnableMargin(margin, !enabled);
  handle(this, MKUINT(0, SEL_FOCUS_SELF), ptr);
  return 1;
}


long FXStyledText::onUpdMargin(FXObject* obj, FXSelector sel, void* ptr) {
  FXint id = SELID(sel);
	FXint margin;
	switch (id) {
	  case ID_LINENUM_MARGIN: margin = 0; break;
	  case ID_SYMBOL_MARGIN:  margin = 1; break;
    case ID_FOLD_MARGIN:    margin = 2; break;
	}
  FXbool enabled = IsEnabledMargin(margin);
	if (enabled) obj->handle(this,MKUINT(ID_CHECK,SEL_COMMAND),NULL);
	else obj->handle(this,MKUINT(ID_UNCHECK,SEL_COMMAND),NULL);
  return 1;
}


///////////////////////////////////////////////////////////////////////////////
//// Clipboard

long FXStyledText::onCmdCutSel(FXObject*, FXSelector, void* ptr) {
  CutSel();
  handle(this, MKUINT(0, SEL_FOCUS_SELF), ptr);
  return 1;
}


long FXStyledText::onCmdCopySel(FXObject*, FXSelector, void* ptr) {
  CopySel();
  handle(this, MKUINT(0, SEL_FOCUS_SELF), ptr);
  return 1;
}


long FXStyledText::onCmdPasteSel(FXObject*, FXSelector, void* ptr) {
  //if(isEditable()) {
    PasteSel();
    handle(this, MKUINT(0, SEL_FOCUS_SELF), ptr);
  //}
  return 1;
}


long FXStyledText::onCmdDeleteSel(FXObject*, FXSelector, void* ptr) {
  //if(isEditable()) {
    ClearSelection();
    handle(this, MKUINT(0, SEL_FOCUS_SELF), ptr);
  //}
  return 1;
}


long FXStyledText::onCmdSelectAll(FXObject*, FXSelector, void*){
  SelectAll();
  return 1;
}


// We now really do have the Clipboard selection
long FXStyledText::onClipboardGained(FXObject* sender, FXSelector sel, void* ptr){
  FXScrollArea::onClipboardGained(sender, sel, ptr);
  return 1;
}


// We lost the Clipboard selection somehow
long FXStyledText::onClipboardLost(FXObject* sender, FXSelector sel, void* ptr){
  FXScrollArea::onClipboardLost(sender ,sel, ptr);
  FXFREE(&clipbuffer);
  clipbuffer = NULL;
  cliplength = 0;
  return 1;
}


// Somebody wants our Clipboard selection
long FXStyledText::onClipboardRequest(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent *event=(FXEvent*)ptr;

  // Try handling it in base class first
  if(FXScrollArea::onClipboardRequest(sender,sel,ptr)) return 1;

  // Requested data from clipboard
  if(event->target == stringType) {
    FXASSERT(cliplength);
    FXuchar *data;
#ifndef WIN32
    FXMALLOC(&data, FXuchar, cliplength);
    memcpy(data, clipbuffer, cliplength);
    setDNDData(FROM_CLIPBOARD, stringType, data, cliplength);
#else
    FXint i, j, c;
    FXCALLOC(&data, FXuchar, 2*cliplength+1);
    for(i=j=0; j<cliplength; j++) {
      c = clipbuffer[j];
      if(c=='\n'){data[i++]='\r';data[i++]='\n';}else{data[i++]=c;}
    }
    setDNDData(FROM_CLIPBOARD, stringType, data, i+1);
#endif

    return 1;
  }

  return 0;
}


///////////////////////////////////////////////////////////////////////////////
//// X Selection

// We now really do have the X selection
long FXStyledText::onSelectionGained(FXObject* sender, FXSelector sel, void* ptr){
//printf("onSelectionGained() entered.\n");
  FXScrollArea::onSelectionGained(sender, sel, ptr);
  return 1;
}


// We lost the X selection somehow
long FXStyledText::onSelectionLost(FXObject* sender, FXSelector sel, void* ptr){
//printf("onSelectionLost() entered.\n");
  FXScrollArea::onSelectionLost(sender, sel, ptr);
  return 1;
}


// Somebody wants our X selection
long FXStyledText::onSelectionRequest(FXObject* sender, FXSelector sel, void* ptr){
//printf("onSelectionRequest() entered.\n");

  // Perhaps the target wants to supply its own data for the selection
  if(FXScrollArea::onSelectionRequest(sender,sel,ptr)) return 1;

  FXEvent *event=(FXEvent*)ptr;

  // Return text of the selection
  if(event->target == stringType){
    FXuchar *data;
    FXint start, end, len;
    GetSelection(&start, &end);
    FXASSERT(start <= end);
    len = end - start;
    FXMALLOC(&data, FXuchar, len);
    memcpy((FXchar*)data, GetSelectedText().text(), len);
    setDNDData(FROM_SELECTION, stringType, data, len);
    return 1;
  }

  return 0;
}


void FXStyledText::grabSelection() {
  FXint what[2];
  FXint start=-1, end=-1, anchor=-1, pos=-1;
  GetSelection(&start, &end);

  if (start > end) return;

  if ((start == end) && hasSelection()) {
    if(target){
      what[0]=start;        //??
      what[1]=end-start;    //??
      target->handle(this,MKUINT(message,SEL_DESELECTED),(void*)what);
    }
    releaseSelection();
    return;
  }

  if ((start < end) /*&& !hasSelection()*/) {
    if(target){
      what[0]=start;        //??
      what[1]=end-start;    //??
      target->handle(this,MKUINT(message,SEL_SELECTED),(void*)what);
    }
    FXDragType types[1];
    types[0] = stringType;
    acquireSelection(types, 1);
  }
}

///////////////////////////////////////////////////////////////////////////////
