// stc.h - Styled Text Control using the Scintilla cource code edit control
// implemented for the FOX toolkit
// SB: initial release; based on ScintillaWX by Robin Dunn.

#ifndef __stc_h__
#define __stc_h__


#include "ScintillaFOX.h"


class  WordList;
struct SCNotification;


/// Multiline STYLED text widget

class FXAPI FXStyledText : public FXScrollArea {
  FXDECLARE(FXStyledText)

protected:

  FXFont        *font;                // Text font
  FXColor       textColor;            // Normal text color
  FXCharset     delimiters;           // Delimiters
  // not used yet
  //FXint       margintop;            // Margins top
  //FXint       marginbottom;         // Margin bottom
  //FXint       marginleft;           // Margin left
  //FXint       marginright;          // Margin right
  FXuint        mode;                 // Mode widget is in

  // in ScintillaFOX:
  FXchar        *clipbuffer;          // Clipped text
  FXint         cliplength;           // Length of clipped text

  ScintillaFX   *m_sfx;               // THE scintilla widget

  FXint         m_itemHeight;         // line height in pixs
  FXint         m_nitems;             // number of lines in the document

  FXTimer*      blinker;
  FXint         m_BlinkSpeed;
	FXint         m_MarginSize[3];      // in chars
	FXString      m_dragtext;           // dragged text
  FXCursor*     m_arrow1;
  FXCursor*     m_arrow2;
  FXCursor*     m_arrow3;
  FXCursor*     m_arrow4;
  FXMenuPane*   m_pane;               //popup


protected:
  FXStyledText();
  static inline FXint fxabs(FXint a){ return a<0?-a:a; }


private:
  FXStyledText(const FXStyledText&);
  FXStyledText& operator=(const FXStyledText&);
  long SendMsg(int msg, long wp=0, long lp=0);  // For sending commands to Scintilla

  enum {
    MOUSE_NONE,
    MOUSE_CHARS,
    MOUSE_WORDS,
    MOUSE_LINES,
    MOUSE_SCROLL,
    MOUSE_DRAG,
    MOUSE_PASTE,
    MOUSE_TRYDRAG
  };


public:
  // System messages
  long onPaint            (FXObject*, FXSelector, void*);
  long onUpdate           (FXObject*, FXSelector, void*);
  long onSize             (FXObject*, FXSelector, void*);
  long onKeyPress         (FXObject*, FXSelector, void*);
  long onKeyRelease       (FXObject*, FXSelector, void*);
  long onFocusIn          (FXObject*, FXSelector, void*);
  long onFocusOut         (FXObject*, FXSelector, void*);
  long onBlink            (FXObject*, FXSelector, void*);
  long onLeftBtnPress     (FXObject*, FXSelector, void*);
  long onLeftBtnRelease   (FXObject*, FXSelector, void*);
  long onRightBtnPress    (FXObject*, FXSelector, void*);
  long onRightBtnRelease  (FXObject*, FXSelector, void*);
  long onMiddleBtnPress   (FXObject*, FXSelector, void*);
  long onMiddleBtnRelease (FXObject*, FXSelector, void*);
  long onMotion           (FXObject*, FXSelector, void*);
	long onBeginDrag        (FXObject*, FXSelector, void*);
	long onDragged          (FXObject*, FXSelector, void*);
  long onEndDrag          (FXObject*, FXSelector, void*);
	long onDNDEnter         (FXObject*, FXSelector, void*);
	long onDNDLeave         (FXObject*, FXSelector, void*);
	long onDNDMotion        (FXObject*, FXSelector, void*);
  long onDNDDrop          (FXObject*, FXSelector, void*);
	long onDNDRequest       (FXObject*, FXSelector, void*);

	// Editing commands
  long onCmdInsertChar   (FXObject*, FXSelector, void*);
  long onCmdInsertString (FXObject*, FXSelector, void*);
  long onCmdGotoLine     (FXObject*, FXSelector, void*);
	long onCmdUndo         (FXObject*, FXSelector, void*);
	long onCmdRedo         (FXObject*, FXSelector, void*);

  // Margins
	long onCmdTogleMargin  (FXObject*, FXSelector, void*);
	long onUpdMargin       (FXObject*, FXSelector, void*);

  // Cut/Copy/Paste/Delete/SelectAll
  //long onUpdHaveSelection(FXObject*, FXSelector, void*);  //todo
  //long onUpdYes(FXObject*, FXSelector, void*);            //todo
  long onCmdCutSel    (FXObject*, FXSelector, void*);
  long onCmdCopySel   (FXObject*, FXSelector, void*);
  long onCmdPasteSel  (FXObject*, FXSelector, void*);
  long onCmdDeleteSel (FXObject*, FXSelector, void*);
  long onCmdSelectAll (FXObject*, FXSelector, void*);

  // Clipboard
  long onClipboardLost    (FXObject*, FXSelector, void*);
  long onClipboardGained  (FXObject*, FXSelector, void*);
  long onClipboardRequest (FXObject*, FXSelector, void*);

  // X Selection
  long onSelectionLost    (FXObject*, FXSelector, void*);
  long onSelectionGained  (FXObject*, FXSelector, void*);
  long onSelectionRequest (FXObject*, FXSelector, void*);

  // Scintilla messages
  long onSciChange (FXObject*, FXSelector, void*);
  long onSciNotify (FXObject*, FXSelector, void*);

  // Scrolling
  long onHScrollerChanged (FXObject*, FXSelector, void*);
  long onHScrollerDragged (FXObject*, FXSelector, void*);
  long onVScrollerChanged (FXObject*, FXSelector, void*);
  long onVScrollerDragged (FXObject*, FXSelector, void*);


  // Todo: Verify which should be protected/private:

  // FXScrollArea virtuals
  virtual FXint getContentWidth();
  virtual FXint getContentHeight();
  virtual FXint getViewportWidth();
  virtual FXint getViewportHeight();

  // FXWindow virtuals
  virtual FXbool canFocus() const;
  virtual void enable();
  virtual void disable();
  virtual void recalc();

  long StartBlinker();
  long StopBlinker();

  void MouseDown(FXObject*, FXSelector, void*);
  void MouseUp(FXObject*, FXSelector, void*);
  void MouseMove(FXObject*, FXSelector, void*);

	void EnableMargin(FXint margin, FXbool show);
	FXbool IsEnabledMargin(FXint margin);
	FXint SetMarginSize(FXint margin, FXint chars);
  void SetDNDCursor(void* ptr);

  void grabSelection();
  void LoadConfig();  //temp
  void SetFont(const FXString& face, int size);


//////////////////////////////////////////////////////////////////////////////

public:

  /// Construct multi-line styled text widget
  FXStyledText(FXComposite *p, FXObject* tgt=NULL, FXSelector sel=0,
          FXuint opts=0, FXint x=0, FXint y=0, FXint w=0, FXint h=0);

  /// Create server-side resources
  virtual void create();

  /// Destructor
  virtual ~FXStyledText();


#include "Messages.h"  // The Scintilla functions

enum {
    ID_THIS = FXScrollArea::ID_LAST,
    ID_INSERT_CHAR,
    ID_INSERT_STRING,
    ID_EDIT,	//the scroll area

    ID_CALLTIP,
    ID_AUTOCOMPLETE,
    ID_UNDO,
    ID_REDO,
    ID_CUT_SEL,
    ID_COPY_SEL,
    ID_PASTE_SEL,
    ID_DELETE_SEL,
    ID_SELECTALL,

    ID_GOTO_LINE,
    ID_SEARCH,
    ID_REPLACE,
		ID_LINENUM_MARGIN,
		ID_SYMBOL_MARGIN,
		ID_FOLD_MARGIN,

    ID_LAST
};


  friend class ScintillaFX;
  friend class Platform;
};

#endif
